from posixpath import normcase
import cv2
import os
import numpy as np
import pandas as pd
import scipy.io as spio
from scipy import ndimage as nd
from skimage import draw
from os import listdir
from os.path import isfile, join
import pydicom
import pydicom.pixel_data_handlers.gdcm_handler as gdcm_handler

pydicom.config.image_handlers = [None, gdcm_handler]


def chunk_it(seq, num):
    avg = len(seq) / float(num)
    out = []
    last = 0.0

    while last < len(seq):
        out.append(seq[int(last) : int(last + avg)])
        last += avg

    return out


def loadmat(filename):
    """
    this function should be called instead of direct spio.loadmat
    as it cures the problem of not properly recovering python dictionaries
    from mat files. It calls the function check keys to cure all entries
    which are still mat-objects
    """

    def _check_keys(d):
        """
        checks if entries in dictionary are mat-objects. If yes
        todict is called to change them to nested dictionaries
        """
        for key in d:
            if isinstance(d[key], spio.matlab.mio5_params.mat_struct):
                d[key] = _todict(d[key])
        return d

    def _todict(matobj):
        """
        A recursive function which constructs from matobjects nested dictionaries
        """
        d = {}
        for strg in matobj._fieldnames:
            elem = matobj.__dict__[strg]
            if isinstance(elem, spio.matlab.mio5_params.mat_struct):
                d[strg] = _todict(elem)
            elif isinstance(elem, np.ndarray):
                d[strg] = _tolist(elem)
            else:
                d[strg] = elem
        return d

    def _tolist(ndarray):
        """
        A recursive function which constructs lists from cellarrays
        (which are loaded as numpy ndarrays), recursing into the elements
        if they contain matobjects.
        """
        elem_list = []
        for sub_elem in ndarray:
            if isinstance(sub_elem, spio.matlab.mio5_params.mat_struct):
                elem_list.append(_todict(sub_elem))
            elif isinstance(sub_elem, np.ndarray):
                elem_list.append(_tolist(sub_elem))
            else:
                elem_list.append(sub_elem)
        return elem_list

    data = spio.loadmat(filename, struct_as_record=False, squeeze_me=True)
    del data["__header__"]
    del data["__version__"]
    del data["__globals__"]
    return _check_keys(data)


def get_scan(scan_path, geno, idx):

    curr_scan_dir = geno["Scan"][idx]
    curr_meta = geno["ScanMetadata"][idx]
    curr_info = geno["IVDInfo"][idx]
    curr_quad = geno["BoundingQ"][idx]

    if isinstance(curr_meta, (list,)):
        pixel_spacing = curr_meta[0]["PixelSpacing"][0]
    else:
        pixel_spacing = curr_meta["PixelSpacing"][0]

    mid_sag = np.round(np.mean(np.array(curr_info["midSagittal"]) - 1.0)).astype(
        np.uint8
    )

    # Slice index
    slice_idx = mid_sag

    # Get slice from volume
    if isinstance(curr_scan_dir, (list,)):
        s_temp = pydicom.dcmread(
            curr_scan_dir[slice_idx].replace("../../Dataset/", scan_path), force=True
        )
        if len(s_temp.file_meta) == 0:
            s_temp.file_meta.TransferSyntaxUID = pydicom.uid.ImplicitVRLittleEndian
        try:
            scan = np.array(s_temp.pixel_array)
        except:
            curr_scan_dir = curr_scan_dir[0]
            slash_idx = [m.start() for m in re.finditer("/", curr_scan_dir)]
            scan = np.array(
                sio.loadmat(
                    "genomat/"
                    + curr_scan_dir[slash_idx[3] + 1 : slash_idx[-1]]
                    + "/scan.mat",
                    verify_compressed_data_integrity=False,
                )["scan"]
            )
            scan = scan[:, :, slice_idx]
    else:
        s_temp = pydicom.dcmread(
            curr_scan_dir.replace("../../Dataset/", scan_path), force=True
        )
        if len(s_temp.file_meta) == 0:
            s_temp.file_meta.TransferSyntaxUID = pydicom.uid.ImplicitVRLittleEndian
        scan = np.array(s_temp.pixel_array)
        if np.size(np.shape(scan)) == 3:
            scan = scan[slice_idx, :, :]

    # Normalize
    scan_shape = np.shape(scan)
    scan = scan.astype(float)
    scan = (scan - np.min(scan)) / (np.ptp(scan) + 1e-9)

    # Get bounding quadrilaterals
    if pixel_spacing == 0:
        resize_factor = 1
    else:
        resize_factor = pixel_spacing / 0.5859
    resize_factor = 1 / resize_factor

    try:
        x = np.array(curr_quad[slice_idx]["x"])
        y = np.array(curr_quad[slice_idx]["y"])
    except:
        x = np.array(curr_quad["x"])
        y = np.array(curr_quad["y"])
    x = x * resize_factor
    y = y * resize_factor
    x = np.array(x).astype(float)
    y = np.array(y).astype(float)
    return scan, x, y


def get_patch_ex(scan, x, y):
    # Initial param
    height = scan.shape[0]
    width = scan.shape[1]
    theta = np.degrees(np.arctan2(y[0] - y[3], x[0] - x[3]))

    # Rotate
    scanrot, qx, qy, offset_w, offset_h = rotate_bb_and_scan(
        scan, x, y, width, height, theta
    )

    # Patch
    pad_add = 5
    w_st = np.round(min(qx))
    w_en = np.round(max(qx))
    h_st = np.round(min(qy))
    h_en = np.round(max(qy))
    w_pad = np.round((w_en - w_st))
    h_pad = np.round((h_en - h_st))
    w_st -= pad_add
    w_en += pad_add
    h_st -= pad_add
    h_en += pad_add
    if h_st < 0:
        h_st = 0
    if w_st < 0:
        w_st = 0
    scan_patch = scanrot[int(h_st) : int(h_en), int(w_st) : int(w_en)]

    return scan_patch


def get_patch(scan, x, y):
    # Initial param
    height = scan.shape[0]
    width = scan.shape[1]
    theta = np.degrees(np.arctan2(y[0] - y[3], x[0] - x[3]))

    # Rotate
    scanrot, qx, qy, offset_w, offset_h = rotate_bb_and_scan(
        scan, x, y, width, height, theta
    )

    # Patch
    pad_mult = 0.75
    w_st = np.round(min(qx))
    w_en = np.round(max(qx))
    h_st = np.round(min(qy))
    h_en = np.round(max(qy))
    w_pad = np.round((w_en - w_st))
    h_pad = np.round((h_en - h_st))
    w_st -= w_pad * pad_mult
    w_en += w_pad * pad_mult
    h_st -= h_pad * pad_mult
    h_en += h_pad * pad_mult
    if h_st < 0:
        h_st = 0
    if w_st < 0:
        w_st = 0
    scan_patch = scanrot[int(h_st) : int(h_en), int(w_st) : int(w_en)]

    new_h = scan_patch.shape[0]
    new_w = scan_patch.shape[1]
    qx_offset = -np.round(min(qx)) + w_pad * pad_mult
    qy_offset = -np.round(min(qy)) + h_pad * pad_mult
    qx = qx + qx_offset
    qy = qy + qy_offset
    qy = qy / new_h * 224
    qx = qx / new_w * 224
    scan_patch = cv2.resize(scan_patch, (224, 224), interpolation=cv2.INTER_CUBIC)

    return scan_patch, new_w, new_h, qx_offset, qy_offset, offset_w, offset_h, theta


def rotate_bb_and_scan(scan, x, y, width, height, theta):
    rotation_matrix = cv2.getRotationMatrix2D((width / 2, height / 2), theta, scale=1)
    cos = np.abs(rotation_matrix[0, 0])
    sin = np.abs(rotation_matrix[0, 1])
    new_w = int((height * sin) + (width * cos))
    new_h = int((height * cos) + (width * sin))
    offset_w = (new_w - width) / 2
    offset_h = (new_h - height) / 2
    rotation_matrix[0, 2] += offset_w
    rotation_matrix[1, 2] += offset_h
    v = np.vstack((x.ravel(), y.ravel(), np.ones(np.size(y.ravel()))))
    calculated = np.dot(rotation_matrix, v)
    scanrot = cv2.warpAffine(
        scan, rotation_matrix, (int(new_w), int(new_h)), flags=cv2.INTER_CUBIC
    )
    qx = calculated[0]
    qy = calculated[1]

    return scanrot, qx, qy, offset_w, offset_h


def rotate_bb(x, y, width, height, theta):
    rotation_matrix = cv2.getRotationMatrix2D((width / 2, height / 2), theta, scale=1)
    cos = np.abs(rotation_matrix[0, 0])
    sin = np.abs(rotation_matrix[0, 1])
    new_w = int((height * sin) + (width * cos))
    new_h = int((height * cos) + (width * sin))
    offset_w = (new_w - width) / 2
    offset_h = (new_h - height) / 2
    rotation_matrix[0, 2] += offset_w
    rotation_matrix[1, 2] += offset_h
    v = np.vstack((x.ravel(), y.ravel(), np.ones(np.size(y.ravel()))))
    calculated = np.dot(rotation_matrix, v)
    qx = calculated[0]
    qy = calculated[1]

    return qx, qy


def retransform_bb(
    height, width, qx, qy, new_w, new_h, qx_offset, qy_offset, offset_w, offset_h, theta
):
    rotx = qx * new_w / 224
    roty = qy * new_h / 224
    rotx = rotx - qx_offset
    roty = roty - qy_offset
    rotx = rotx - offset_w
    roty = roty - offset_h
    rotx, roty = rotate_bb(rotx, roty, width, height, -theta)
    rotx = rotx - offset_w
    roty = roty - offset_h

    return rotx, roty


def get_scan_in_list(curr_scan_path, data_path, mat_path):
    # Check if mat files exist
    c_ = curr_scan_path[0]
    c_ = mat_path + "/" + c_[: c_.rfind("/")]
    if os.path.exists(c_):
        # Get volume - mat
        volume = np.array(
            spio.loadmat(c_ + "/scan.mat", verify_compressed_data_integrity=False)[
                "scan"
            ]
        )
    else:
        # Get volume - dicom
        if len(curr_scan_path) == 1:
            # 3D scans
            s_temp = pydicom.dcmread(data_path + "/" + curr_scan_path[0], force=True)
            if len(s_temp.file_meta) == 0:
                s_temp.file_meta.TransferSyntaxUID = pydicom.uid.ImplicitVRLittleEndian
            volume = np.array(s_temp.pixel_array)
            if len(volume.shape) < 3:
                volume = volume[:, :, None]
            elif len(volume.shape) == 3:
                volume = np.transpose(np.array(volume), (1, 2, 0)).astype(float)
            else:
                print("Unknown volume shape")
        else:
            try:
                # 2D scans
                volume = []
                for temp_input in curr_scan_path:
                    # print(temp_input)
                    s_temp = pydicom.dcmread(data_path + "/" + temp_input, force=True)
                    if len(s_temp.file_meta) == 0:
                        s_temp.file_meta.TransferSyntaxUID = (
                            pydicom.uid.ImplicitVRLittleEndian
                        )
                    scan = np.array(s_temp.pixel_array)
                    volume.append(scan)
                volume = np.transpose(np.array(volume), (1, 2, 0)).astype(float)
            except:
                print(temp_input)
    return volume


def get_scan_in_folder(input_path):
    input_files = [f for f in listdir(input_path) if isfile(join(input_path, f))]
    if len(input_files) == 1:
        # 3D scans
        s_temp = pydicom.dcmread(input_path + "/" + input_files[0], force=True)
        if len(s_temp.file_meta) == 0:
            s_temp.file_meta.TransferSyntaxUID = pydicom.uid.ImplicitVRLittleEndian
        volume = np.array(s_temp.pixel_array)
        pixel_spacing = s_temp.PixelSpacing
    else:
        lis = []
        order = []
        for temp_input in input_files:
            s_temp = pydicom.dcmread(input_path + "/" + temp_input, force=True)
            # series_number = int(s_temp.SliceLocation)
            # order.append(series_number)
            series_number = float(s_temp.ImagePositionPatient[0])
            order.append(series_number)
            lis.append(temp_input)
        input_files = [x for _, x in sorted(zip(order, lis))]
        input_files.reverse()
        volume = []
        for temp_input in input_files:
            s_temp = pydicom.dcmread(input_path + "/" + temp_input, force=True)
            if len(s_temp.file_meta) == 0:
                s_temp.file_meta.TransferSyntaxUID = pydicom.uid.ImplicitVRLittleEndian
            scan = np.array(s_temp.pixel_array)
            volume.append(scan)
            pixel_spacing = np.array(s_temp.PixelSpacing)
    volume = np.transpose(np.array(volume), (1, 2, 0)).astype(float)
    return {"volume": volume, "pixel_spacing": pixel_spacing}


def poly2mask(vertex_row_coords, vertex_col_coords, shape):
    fill_row_coords, fill_col_coords = draw.polygon(
        vertex_row_coords, vertex_col_coords, shape
    )
    mask = np.zeros(shape, dtype=bool)
    mask[fill_row_coords, fill_col_coords] = True
    return mask


def get_vbs_intensity(volume, all_vb_x, all_vb_y, all_vb_mid, all_vb_label):
    vbs_intensity = []
    height, width, depth = volume.shape
    for vb_idx in range(all_vb_x.shape[1]):
        curr_label = all_vb_label[vb_idx, :]
        curr_mid = all_vb_mid[vb_idx]
        curr_mask = poly2mask(
            all_vb_y[:, vb_idx, int(curr_mid)],
            all_vb_x[:, vb_idx, int(curr_mid)],
            volume.shape[:2],
        )
        mask = np.tile(curr_mask, (depth, 1, 1))
        mask = np.transpose(mask, (1, 2, 0)).astype(float)
        mask[:, :, ~curr_label.astype(bool)] = 0
        vbs_intensity.append(volume[mask.astype(bool)])
    return vbs_intensity


def rotate_bb_and_volume(volume, x, y, normalise=True):
    height = volume.shape[0]
    width = volume.shape[1]
    depth = volume.shape[2]
    theta = np.degrees(np.arctan2(y[0] - y[3], x[0] - x[3]))

    rotation_matrix = cv2.getRotationMatrix2D((width / 2, height / 2), theta, scale=1)
    cos = np.abs(rotation_matrix[0, 0])
    sin = np.abs(rotation_matrix[0, 1])
    new_w = int((height * sin) + (width * cos))
    new_h = int((height * cos) + (width * sin))
    offset_w = (new_w - width) / 2
    offset_h = (new_h - height) / 2
    rotation_matrix[0, 2] += offset_w
    rotation_matrix[1, 2] += offset_h
    v = np.vstack((x.ravel(), y.ravel(), np.ones(np.size(y.ravel()))))
    calculated = np.dot(rotation_matrix, v)
    qx = calculated[0]
    qy = calculated[1]

    volume_rot = []
    for idx in range(depth):
        if normalise:
            volume_rot.append(
                cv2.warpAffine(
                    volume[:, :, idx],
                    rotation_matrix,
                    (int(new_w), int(new_h)),
                    flags=cv2.INTER_CUBIC,
                )
            )
        else:
            volume_rot.append(
                cv2.warpAffine(
                    volume[:, :, idx],
                    rotation_matrix,
                    (int(new_w), int(new_h)),
                    flags=cv2.INTER_NEAREST,
                )
            )
    volume_rot = np.transpose(np.array(volume_rot), (1, 2, 0)).astype(float)

    min_x = min(qx)
    max_x = max(qx)
    qx[0] = max_x
    qx[1] = max_x
    qx[2] = min_x
    qx[3] = min_x

    min_y = min(qy)
    max_y = max(qy)
    qy[0] = min_y
    qy[1] = max_y
    qy[2] = max_y
    qy[3] = min_y

    return volume_rot, qx, qy


def get_ivd_vol(
    volume, ivd_curr_x, ivd_curr_y, vb_pair_median, curr_ivd_mid, norm_med, patch_size
):
    x = np.array(ivd_curr_x)
    y = np.array(ivd_curr_y)

    # Rotate
    volume_rot, qx, qy = rotate_bb_and_volume(volume, x, y)

    # Add 50% width
    w = max(qx) - min(qx)
    qx[0] += w * 0.5
    qx[1] += w * 0.5
    qx[2] -= w * 0.5
    qx[3] -= w * 0.5
    curr_w = max(qx) - min(qx)
    curr_h = max(qy) - min(qy)
    extra_vert = (curr_w / 2 - curr_h) * 0.5
    qy[0] -= extra_vert
    qy[1] += extra_vert
    qy[2] += extra_vert
    qy[3] -= extra_vert
    curr_w = max(qx) - min(qx)
    curr_h = max(qy) - min(qy)

    #  Jitter Crop: Border = [39.25 46.5]
    extra_w = ((patch_size[1] - 227.0) / 2.0) * (curr_w / 227.0)
    extra_h = ((patch_size[0] - (227.0 / 2.0)) / 2.0) * (curr_h / (227.0 / 2.0))
    min_x = np.round(min(qx) - extra_w)
    max_x = np.round(max(qx) + extra_w)
    min_y = np.round(min(qy) - extra_h)
    max_y = np.round(max(qy) + extra_h)
    curr_w = max_x - min_x
    curr_h = max_y - min_y

    if min_y < 0:
        y_offset = np.round(abs(min_y))
        volume_rot = np.concatenate(
            (
                np.zeros(
                    (y_offset.astype(int), volume_rot.shape[1], volume_rot.shape[2])
                ),
                volume_rot,
            ),
            axis=0,
        )
        min_y += y_offset
        max_y += y_offset
    if max_y >= volume_rot.shape[0]:
        y_offset = (abs(max_y) + 1) - volume_rot.shape[0]
        volume_rot = np.concatenate(
            (
                volume_rot,
                np.zeros((y_offset, volume_rot.shape[1], volume_rot.shape[2])),
            ),
            axis=0,
        )
    if min_x < 0:
        x_offset = np.round(abs(min_x))
        volume_rot = np.concatenate(
            (
                np.zeros(
                    (volume_rot.shape[0], x_offset.astype(int), volume_rot.shape[2])
                ),
                volume_rot,
            ),
            axis=1,
        )
        min_x += x_offset
        max_x += x_offset
    if max_x >= volume_rot.shape[1]:
        x_offset = (abs(max_x) + 1) - volume_rot.shape[1]
        volume_rot = np.concatenate(
            (
                volume_rot,
                np.zeros((volume_rot.shape[0], x_offset, volume_rot.shape[2])),
            ),
            axis=1,
        )

    min_x = np.round(min_x).astype(int)
    max_x = np.round(max_x).astype(int)
    min_y = np.round(min_y).astype(int)
    max_y = np.round(max_y).astype(int)
    sub_vol = volume_rot[min_y:max_y, min_x:max_x, :]

    # Resize
    ivd_vol = []
    for idx in range(sub_vol.shape[2]):
        ivd_vol.append(
            cv2.resize(
                sub_vol[:, :, idx], patch_size[::-1], interpolation=cv2.INTER_CUBIC
            )
        )
    ivd_vol = np.transpose(np.array(ivd_vol), (1, 2, 0)).astype(float)

    # Normalize
    ivd_vol = ivd_vol / vb_pair_median * norm_med
    ivd_vol[ivd_vol < 0] = 0
    ivd_vol[ivd_vol > 2.0] = 2.0
    ivd_vol /= 2.0
    return ivd_vol


def get_all_ivd_vol(volume, all_vb_x, all_vb_y, all_vb_mid, all_vb_label):
    # VBs intensity values
    vbs_intensity = get_vbs_intensity(
        volume, all_vb_x, all_vb_y, all_vb_mid, all_vb_label
    )

    # Get volumes
    ivds = []
    norm_med = 0.5
    patch_size = (192, 320)
    no_of_ivd = len(vbs_intensity) - 1
    for ivd_idx in range(no_of_ivd):
        curr_ivd_mid = np.round(
            np.mean([all_vb_mid[ivd_idx], all_vb_mid[ivd_idx + 1]])
        ).astype(int)
        vb_pair_intensity = np.concatenate(
            (vbs_intensity[ivd_idx], vbs_intensity[ivd_idx + 1])
        )
        vb_pair_median = np.median(vb_pair_intensity)

        vb_curr_x = all_vb_x[:, ivd_idx, curr_ivd_mid]
        vb_curr_y = all_vb_y[:, ivd_idx, curr_ivd_mid]
        vb_next_x = all_vb_x[:, ivd_idx + 1, curr_ivd_mid]
        vb_next_y = all_vb_y[:, ivd_idx + 1, curr_ivd_mid]
        ivd_curr_x = [vb_next_x[1], vb_curr_x[0], vb_curr_x[3], vb_next_x[2]]
        ivd_curr_y = [vb_next_y[1], vb_curr_y[0], vb_curr_y[3], vb_next_y[2]]

        ivd_vol = get_ivd_vol(
            volume,
            ivd_curr_x,
            ivd_curr_y,
            vb_pair_median,
            curr_ivd_mid,
            norm_med,
            patch_size,
        )

        # Centered & choose 15 slices; TO-DO: resize 3D
        ivd_vol_centered = []
        for j in range(-7, 7 + 1, 1):
            curr_slice = curr_ivd_mid + j
            if (curr_slice < 0) | (curr_slice >= ivd_vol.shape[2]):
                temp_IVD = np.zeros(patch_size)
            else:
                temp_IVD = ivd_vol[:, :, curr_slice]
            ivd_vol_centered.append(temp_IVD)
        ivd_vol_centered = np.transpose(np.array(ivd_vol_centered), (1, 2, 0)).astype(
            float
        )
        num_rows, num_cols, num_slices = ivd_vol_centered.shape
        max_cols = num_cols - 48
        min_cols = 48
        max_rows = num_rows - 40
        min_rows = 40
        ivd_vol_centered = ivd_vol_centered[min_rows:max_rows, min_cols:max_cols, 3:12]
        ivd_vol_centered = np.transpose(ivd_vol_centered, (2, 0, 1))
        ivds.append(ivd_vol_centered)
    ivds = np.array(ivds)
    return ivds


def get_vu_vol(
    volume,
    ivd_curr_x,
    ivd_curr_y,
    vb_pair_median,
    curr_ivd_mid,
    norm_med,
    patch_size,
    normalise=True,
):
    x = np.array(ivd_curr_x)
    y = np.array(ivd_curr_y)

    # Rotate
    volume_rot, qx, qy = rotate_bb_and_volume(volume, x, y, normalise=normalise)

    min_x = np.round(min(qx))
    max_x = np.round(max(qx))
    min_y = np.round(min(qy))
    max_y = np.round(max(qy))
    w = max_x - min_x
    h = max_y - min_y

    if w > h:
        max_y += (w - h) / 2.0
        min_y -= (w - h) / 2.0
        h = max_y - min_y
    else:
        max_x += (h - w) / 2.0
        min_x -= (h - w) / 2.0
        w = max_x - min_x

    # Add 50% width/height
    max_x += w * 0.5
    min_x -= w * 0.5
    max_y += h * 0.5
    min_y -= h * 0.5

    if min_y < 0:
        y_offset = abs(min_y).astype(int)
        volume_rot = np.concatenate(
            (
                np.zeros((y_offset, volume_rot.shape[1], volume_rot.shape[2])),
                volume_rot,
            ),
            axis=0,
        )
        min_y += y_offset
        max_y += y_offset
    if max_y >= volume_rot.shape[0]:
        y_offset = ((abs(max_y) + 1) - volume_rot.shape[0]).astype(int)
        volume_rot = np.concatenate(
            (
                volume_rot,
                np.zeros((y_offset, volume_rot.shape[1], volume_rot.shape[2])),
            ),
            axis=0,
        )
    if min_x < 0:
        x_offset = abs(min_x).astype(int)
        volume_rot = np.concatenate(
            (
                np.zeros((volume_rot.shape[0], x_offset, volume_rot.shape[2])),
                volume_rot,
            ),
            axis=1,
        )
        min_x += x_offset
        max_x += x_offset
    if max_x >= volume_rot.shape[1]:
        x_offset = ((abs(max_x) + 1) - volume_rot.shape[1]).astype(int)
        volume_rot = np.concatenate(
            (
                volume_rot,
                np.zeros((volume_rot.shape[0], x_offset, volume_rot.shape[2])),
            ),
            axis=1,
        )

    min_x = np.round(min_x).astype(int)
    max_x = np.round(max_x).astype(int)
    min_y = np.round(min_y).astype(int)
    max_y = np.round(max_y).astype(int)
    sub_vol = volume_rot[min_y:max_y, min_x:max_x, :]

    resize_factors = (
        15 / sub_vol.shape[2],
        patch_size[0] / sub_vol.shape[0],
        patch_size[1] / sub_vol.shape[1],
    )
    # Resize
    ivd_vol = []
    for idx in range(sub_vol.shape[2]):
        if normalise:
            ivd_vol.append(
                cv2.resize(
                    sub_vol[:, :, idx], patch_size[::-1], interpolation=cv2.INTER_CUBIC
                )
            )
        else:
            ivd_vol.append(
                cv2.resize(
                    sub_vol[:, :, idx],
                    patch_size[::-1],
                    interpolation=cv2.INTER_NEAREST,
                )
            )
    ivd_vol = np.transpose(np.array(ivd_vol), (1, 2, 0)).astype(float)

    # Normalize
    if normalise:
        ivd_vol = ivd_vol / vb_pair_median * norm_med
        ivd_vol[ivd_vol < 0] = 0
        ivd_vol[ivd_vol > 2.0] = 2.0
        ivd_vol /= 2.0
    return ivd_vol, resize_factors


def get_all_vus_vol(
    volume, all_vb_x, all_vb_y, all_vb_mid, all_vb_label, normalise=True
):
    # VBs intensity values
    vbs_intensity = get_vbs_intensity(
        volume, all_vb_x, all_vb_y, all_vb_mid, all_vb_label
    )

    # Get IVD volumes
    vus = []
    all_resize_factors = []
    norm_med = 0.5
    patch_size = (320, 320)
    no_of_ivd = len(vbs_intensity)
    for ivd_idx in range(no_of_ivd):
        if ivd_idx != (no_of_ivd - 1):
            curr_ivd_mid = np.round(
                np.mean([all_vb_mid[ivd_idx], all_vb_mid[ivd_idx + 1]])
            ).astype(int)
            vb_pair_intensity = np.concatenate(
                (vbs_intensity[ivd_idx], vbs_intensity[ivd_idx + 1])
            )
            vb_pair_median = np.median(vb_pair_intensity)

            vb_curr_x = all_vb_x[:, ivd_idx, curr_ivd_mid]
            vb_curr_y = all_vb_y[:, ivd_idx, curr_ivd_mid]
            vb_next_x = all_vb_x[:, ivd_idx + 1, curr_ivd_mid]
            vb_next_y = all_vb_y[:, ivd_idx + 1, curr_ivd_mid]

            vu_curr_x = [
                np.mean([vb_next_x[0], vb_next_x[1]]),
                np.mean([vb_curr_x[0], vb_curr_x[1]]),
                np.mean([vb_curr_x[2], vb_curr_x[3]]),
                np.mean([vb_next_x[2], vb_next_x[3]]),
            ]
            vu_curr_y = [
                np.mean([vb_next_y[0], vb_next_y[1]]),
                np.mean([vb_curr_y[0], vb_curr_y[1]]),
                np.mean([vb_curr_y[2], vb_curr_y[3]]),
                np.mean([vb_next_y[2], vb_next_y[3]]),
            ]

        else:
            curr_ivd_mid = np.round(all_vb_mid[ivd_idx]).astype(int)
            vb_pair_median = np.median(vbs_intensity[ivd_idx])

            vb_curr_x = all_vb_x[:, ivd_idx, curr_ivd_mid]
            vb_curr_y = all_vb_y[:, ivd_idx, curr_ivd_mid]

            x0 = vb_curr_x[0]
            y0 = vb_curr_y[0]
            x1 = np.mean([vb_curr_x[0], vb_curr_x[1]])
            y1 = np.mean([vb_curr_y[0], vb_curr_y[1]])
            x2 = np.mean([vb_curr_x[2], vb_curr_x[3]])
            y2 = np.mean([vb_curr_y[2], vb_curr_y[3]])
            x3 = vb_curr_x[3]
            y3 = vb_curr_y[3]
            x0 = x0 + 2 * (x0 - x1)
            y0 = y0 + 2 * (y0 - y1)
            x3 = x3 + 2 * (x3 - x2)
            y3 = y3 + 2 * (y3 - y2)

            vu_curr_x = [x0, x1, x2, x3]
            vu_curr_y = [y0, y1, y2, y3]

        vu_vol, resize_factors = get_vu_vol(
            volume,
            vu_curr_x,
            vu_curr_y,
            vb_pair_median,
            curr_ivd_mid,
            norm_med,
            patch_size,
            normalise=normalise,
        )
        all_resize_factors.append(resize_factors)

        # Resize to 15 slices
        ori_depth = vu_vol.shape[2]
        dsfactor = [
            w / float(f)
            for w, f in zip([vu_vol.shape[0], vu_vol.shape[1], 15], vu_vol.shape)
        ]
        vu_vol = nd.interpolation.zoom(vu_vol, dsfactor)
        curr_ivd_mid = np.round((curr_ivd_mid / ori_depth) * 15).astype(int)

        # Centered & choose 15 slices
        vu_vol_centered = []
        for j in range(-7, 7 + 1, 1):
            curr_slice = curr_ivd_mid + j
            if (curr_slice < 0) | (curr_slice >= vu_vol.shape[2]):
                temp_vu = np.zeros(patch_size)
            else:
                temp_vu = vu_vol[:, :, curr_slice]
            vu_vol_centered.append(temp_vu)
        vu_vol_centered = np.transpose(np.array(vu_vol_centered), (1, 2, 0)).astype(
            float
        )
        vus.append(vu_vol_centered)
    vus = np.array(vus)
    return vus, all_resize_factors


def vert_dicts_to_classification_format(vert_dicts, no_slices):
    """Converts detection output of VFR detector to format that can be read by the classification pipeline"""
    vert_dicts = sorted(
        vert_dicts, key=lambda x: np.array(x["average_polygon"])[:, 1].mean(), reverse=True
    )
    all_vb_x = np.zeros((4, len(vert_dicts), no_slices))
    all_vb_y = np.zeros((4, len(vert_dicts), no_slices))
    all_vb_label = np.zeros((len(vert_dicts), no_slices))
    all_vb_mid = np.zeros(len(vert_dicts))
    for vert_dict_idx, vert_dict in enumerate(vert_dicts):
        # begin by filling all entries with the average polygon
        all_vb_x[:, vert_dict_idx, :] = np.stack(
            [np.array(vert_dict["average_polygon"])[:, 0]] * no_slices, axis=-1
        )
        all_vb_y[:, vert_dict_idx, :] = np.stack(
            [np.array(vert_dict["average_polygon"])[:, 1]] * no_slices, axis=-1
        )
        for poly_idx, poly in enumerate(vert_dict["polys"]):
            slice_no = vert_dict["slice_nos"][poly_idx]
            poly = np.array(poly)
            all_vb_x[:, vert_dict_idx, slice_no] = poly[:, 0]
            all_vb_y[:, vert_dict_idx, slice_no] = poly[:, 1]
            all_vb_label[vert_dict_idx, slice_no] = 1

        all_vb_mid[vert_dict_idx] = np.round(np.median(vert_dict["slice_nos"]))
    vb_level_names = [vert_dict["predicted_label"] for vert_dict in vert_dicts]

    return all_vb_x, all_vb_y, all_vb_mid, all_vb_label, vb_level_names


def get_ivd_level_names(vu_level_names):
    ivd_level_names = []
    for idx in range(len(vu_level_names)):
        if idx == 0:
            continue
        ivd_level_names.append(f"{vu_level_names[idx]}-{vu_level_names[idx-1]}")
    return ivd_level_names


def format_gradings(gradings, ivd_level_names):
    ivd_level_names.reverse()
    col = []
    zl = []
    for key in gradings:
        gradings[key] = np.flip(gradings[key])
        if key in ["Pfirrmann", "Narrowing", "CentralCanalStenosis"]:
            gradings[key] += 1
        col.append(key)
        zl.append(gradings[key])
    gradings = pd.DataFrame(list(zip(*zl)), index=ivd_level_names, columns=col)
    return gradings



from .appearance import AppearanceModel
from .context import ContextModel
from .vfr import VFRResNetDetector
from .grading import GradingModel

__all__ = [
    "AppearanceModel",
    "ContextModel",
    "VFRResNetDetector",
    "GradingModel",
]




import pathlib
import sys
sys.path.insert(0, pathlib.Path(__file__).absolute().parents[1].resolve().as_posix())


# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'SpineNet'
copyright = '2022, Rhydian Windsor'
author = 'Rhydian Windsor'
release = 'v2.0.0'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    'sphinx.ext.doctest',
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.napoleon'
]

templates_path = ['_templates']
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']



# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'alabaster'
html_static_path = ['_static']



import numpy as np
from numpy import polyfit, polyval


def discard_outliers(vert_dicts, polyfit_degree=5):
    """
    Fits curve through detected vertebra and discards outliers that are far from curve

    This is done by using a quintic curve in the sagittal plane, fit to the vertebra
    centroids. If no part of the curve passes through a detection box, it is discarded

    Inputs
    -------
    vert_dicts : list of dicts
        output of utils/detect_and group function. List of dictionaries where each element
        describes a different vertebra detected

    Outputs
    -------
    vert_dicts : list of dicts
        same as input but with outlier vertebrae removed
    """

    top_points = np.zeros((len(vert_dicts), 2))
    bottom_points = np.zeros((len(vert_dicts), 2))
    for vert_idx, vert in enumerate(vert_dicts):
        top_points[vert_idx, :] = np.mean(
            np.take(vert["average_polygon"], [0, 3], axis=0), axis=0
        )
        bottom_points[vert_idx, :] = np.mean(
            np.take(vert["average_polygon"], [1, 2], axis=0), axis=0
        )
        # top_points[vert_idx, :] = np.mean(vert['average_polygon'], axis=0)

    points = np.concatenate((top_points, bottom_points), axis=0)
    xs = points[:, 1]
    ys = points[:, 0]
    if len(xs) == 0:
        # if there are no vertebrae, quit function here
        return vert_dicts, None
    cooefs = polyfit(xs, ys, deg=polyfit_degree)
    distances_from_curve = np.array(
        [np.abs(polyval(cooefs, xs[i]) - ys[i]) for i in range(len(vert_dicts))]
    )

    dropped_indices = []
    # for each vert check the curve passes through them. If not, discard them
    for vert_idx, vert in enumerate(vert_dicts):
        average_polygon = vert["average_polygon"]
        passes_top = False
        passes_bottom = False
        # check top
        if average_polygon[0, 0] > polyval(
            cooefs, average_polygon[0, 1]
        ) and average_polygon[-1, 0] < polyval(cooefs, average_polygon[-1, 1]):
            passes_top = True
        # check bottom
        if average_polygon[1, 0] > polyval(
            cooefs, average_polygon[1, 1]
        ) and average_polygon[2, 0] < polyval(cooefs, average_polygon[2, 1]):
            passes_bottom = True
        if passes_top or passes_bottom:
            continue
        else:
            dropped_indices.append(vert_idx)

    dropped_indices.sort(reverse=True)
    for dropped_index in dropped_indices:
        vert_dicts.pop(dropped_index)
    # fit curve to the points
    return vert_dicts



# VFR
from .gen_utils import *
from .scan_preprocessing import split_into_patches_exhaustive
from .detection_post_processing import make_in_slice_detections
from .detect_and_group import detect_and_group
from .discard_outliers import discard_outliers
from .extract_volumes import extract_volumes
from .label_verts import label_verts
from .gt_formatting import *

# Classification
from .classification import classify_ivd, classify_ivd_v2_resnet



import os
import glob
import torch
import torch.nn as nn
from collections.abc import Iterable, Iterator
from typing import Tuple, List


def conv3x3(in_planes, out_planes, stride=1, groups=1, dilation=1):
    """3x3 convolution with padding"""
    return nn.Conv3d(
        in_planes,
        out_planes,
        (3, 3, 3),
        stride=(1, stride, stride),
        padding=(1, dilation, dilation),
        groups=groups,
        bias=False,
        dilation=dilation,
    )


def conv1x1(in_planes, out_planes, stride=1):
    """1x1 convolution"""
    return nn.Conv3d(
        in_planes,
        out_planes,
        (1, 1, 1),
        stride=(1, stride, stride),
        padding=(0, 0, 0),
        bias=False,
    )


class BasicBlock(nn.Module):
    expansion = 1

    def __init__(
        self,
        inplanes,
        planes,
        stride=1,
        downsample=None,
        groups=1,
        base_width=64,
        dilation=1,
        norm_layer=None,
    ) -> None:
        super(BasicBlock, self).__init__()
        if norm_layer is None:
            norm_layer = nn.BatchNorm3d
        if groups != 1 or base_width != 64:
            raise ValueError("BasicBlock only supports groups=1 and base_width=64")
        if dilation > 1:
            raise NotImplementedError("Dilation > 1 not supported in BasicBlock")
        # Both self.conv1 and self.downsample layers downsample the input when stride != 1
        self.conv1 = conv3x3(inplanes, planes, stride)
        self.bn1 = norm_layer(planes)
        self.relu = nn.ReLU(inplace=True)
        self.conv2 = conv3x3(planes, planes)
        self.bn2 = norm_layer(planes)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        identity = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        out += identity
        out = self.relu(out)

        return out


class Bottleneck(nn.Module):
    # Bottleneck in torchvision places the stride for downsampling at 3x3 convolution(self.conv2)
    # while original implementation places the stride at the first 1x1 convolution(self.conv1)
    # according to "Deep residual learning for image recognition"https://arxiv.org/abs/1512.03385.
    # This variant is also known as ResNet V1.5 and improves accuracy according to
    # https://ngc.nvidia.com/catalog/model-scripts/nvidia:resnet_50_v1_5_for_pytorch.

    expansion = 4

    def __init__(
        self,
        inplanes,
        planes,
        stride=1,
        downsample=None,
        groups=1,
        base_width=64,
        dilation=1,
        norm_layer=None,
    ):
        super(Bottleneck, self).__init__()
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        width = int(planes * (base_width / 64.0)) * groups
        # Both self.conv2 and self.downsample layers downsample the input when stride != 1
        self.conv1 = conv1x1(inplanes, width)
        self.bn1 = norm_layer(width)
        self.conv2 = conv3x3(width, width, stride, groups, dilation)
        self.bn2 = norm_layer(width)
        self.conv3 = conv1x1(width, planes * self.expansion)
        self.bn3 = norm_layer(planes * self.expansion)
        self.relu = nn.ReLU(inplace=True)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        identity = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        out += identity
        out = self.relu(out)

        return out


class GradingModel(nn.Module):
    def __init__(
        self,
        block=BasicBlock,
        layers: List[int] = [3, 4, 6, 3],
        num_classes: int = 2,
        zero_init_residual: bool = False,
        groups: int = 1,
        width_per_group=64,
        norm_layer=None,
    ) -> None:
        super().__init__()
        if norm_layer is None:
            norm_layer = nn.BatchNorm3d
        self._norm_layer = norm_layer

        self.inplanes = 64
        self.dilation = 1
        replace_stride_with_dilation = [False, False, False]
        self.groups = groups
        self.base_width = width_per_group
        self.conv1 = nn.Conv3d(
            1, self.inplanes, (3, 7, 7), stride=(1, 2, 2), padding=(1, 3, 3)
        )
        self.bn1 = norm_layer(self.inplanes)
        self.relu = nn.ReLU(inplace=True)
        self.maxpool = nn.MaxPool3d((1, 3, 3), stride=(1, 2, 2), padding=(0, 1, 1))
        self.layer1 = self._make_layer(block, 64, layers[0])
        # self.layer2 = self._make_layer(block, 128, layers[1], stride=2,
        #                                dilate=replace_stride_with_dilation[0])
        self.layer2 = self._make_layer(
            block, 128, layers[1], stride=1, dilate=replace_stride_with_dilation[0]
        )
        self.layer3 = self._make_layer(
            block, 256, layers[2], stride=2, dilate=replace_stride_with_dilation[1]
        )
        self.layer4 = self._make_layer(
            block, 512, layers[3], stride=2, dilate=replace_stride_with_dilation[2]
        )
        self.avgpool = nn.AdaptiveAvgPool3d((1, 1, 1))
        self.fc_pf = nn.Linear(512 * block.expansion, 5)
        self.fc_nar = nn.Linear(512 * block.expansion, 4)
        self.fc_ccs = nn.Linear(512 * block.expansion, 4)
        self.fc_spn = nn.Linear(512 * block.expansion, 3)
        self.fc_ued = nn.Linear(512 * block.expansion, 2)
        self.fc_led = nn.Linear(512 * block.expansion, 2)
        self.fc_umc = nn.Linear(512 * block.expansion, 2)
        self.fc_lmc = nn.Linear(512 * block.expansion, 2)
        self.fc_fsl = nn.Linear(512 * block.expansion, 2)
        self.fc_fsr = nn.Linear(512 * block.expansion, 2)
        self.fc_hrn = nn.Linear(512 * block.expansion, 2)

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode="fan_out", nonlinearity="relu")
            elif isinstance(m, (nn.BatchNorm2d, nn.GroupNorm)):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)

        # Zero-initialize the last BN in each residual branch,
        # so that the residual branch starts with zeros, and each residual block behaves like an identity.
        # This improves the model by 0.2~0.3% according to https://arxiv.org/abs/1706.02677
        if zero_init_residual:
            for m in self.modules():
                if isinstance(m, Bottleneck):
                    nn.init.constant_(m.bn3.weight, 0)
                elif isinstance(m, BasicBlock):
                    nn.init.constant_(m.bn2.weight, 0)

    def _make_layer(self, block, planes, blocks, stride=1, dilate=False):
        norm_layer = self._norm_layer
        downsample = None
        previous_dilation = self.dilation
        if dilate:
            self.dilation *= stride
            stride = 1
        if stride != 1 or self.inplanes != planes * block.expansion:
            downsample = nn.Sequential(
                conv1x1(self.inplanes, planes * block.expansion, stride),
                norm_layer(planes * block.expansion),
            )

        layers = []
        layers.append(
            block(
                self.inplanes,
                planes,
                stride,
                downsample,
                self.groups,
                self.base_width,
                previous_dilation,
                norm_layer,
            )
        )
        self.inplanes = planes * block.expansion
        for _ in range(1, blocks):
            layers.append(
                block(
                    self.inplanes,
                    planes,
                    groups=self.groups,
                    base_width=self.base_width,
                    dilation=self.dilation,
                    norm_layer=norm_layer,
                )
            )

        return nn.Sequential(*layers)

    def forward(self, x: torch.Tensor) -> Tuple[torch.Tensor]:
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)
        x = self.maxpool(x)

        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)

        x = self.avgpool(x)
        x = torch.flatten(x, 1)
        x_pf = self.fc_pf(x)
        x_nar = self.fc_nar(x)
        x_ccs = self.fc_ccs(x)
        x_spn = self.fc_spn(x)
        x_ued = self.fc_ued(x)
        x_led = self.fc_led(x)
        x_umc = self.fc_umc(x)
        x_lmc = self.fc_lmc(x)
        x_fsl = self.fc_fsl(x)
        x_fsr = self.fc_fsr(x)
        x_hrn = self.fc_hrn(x)

        return (
            x_pf,
            x_nar,
            x_ccs,
            x_spn,
            x_ued,
            x_led,
            x_umc,
            x_lmc,
            x_fsl,
            x_fsr,
            x_hrn,
        )

    def _get_classification_layers(self):
        disease_categories = [
            "pf",
            "nar",
            "ccs",
            "spn",
            "ued",
            "led",
            "umc",
            "lmc",
            "fsl",
            "fsr",
            "hrn",
        ]
        is_classification_layer = (
            lambda x: torch.Tensor(
                [disease_category in x[0] for disease_category in disease_categories]
            )
            .bool()
            .any()
        )
        for child in self.named_children():
            if is_classification_layer(child):
                yield child
            else:
                continue

    def finetune(self, reset_weights: bool = True) -> None:
        """Freeze all layers except classification, and reset weights
        for classification_layer"""
        for parameter in self.parameters():
            parameter.requires_grad = False
        for classification_layer in self._get_classification_layers():
            for parameter in classification_layer[1].parameters():
                parameter.requires_grad = True
            if reset_weights:
                classification_layer[1].reset_parameters()

    def load_weights(self, save_path: str, verbose: bool = True) -> None:
        if os.path.isdir(save_path):
            list_of_pt = glob.glob(save_path + "/*.pt")
            latest_pt = max(list_of_pt, key=os.path.getctime)
            checkpoint = torch.load(latest_pt, map_location="cpu")
            self.load_state_dict(checkpoint["model_weights"])
            start_epoch = checkpoint["epoch_no"] + 1
            if verbose:
                print(f"==> Loading model trained for {start_epoch} epochs...")
        else:
            raise NameError(f"save path {save_path} could not be found")
        return



import numpy as np
import cv2
import torch
from typing import List, Tuple, Dict, Union


def normalize_patch(
        patch: np.ndarray,
        upper_percentile: float = 99.5,
        lower_percentile: float = 0.5
) -> np.ndarray:
    """
    Normalize a single image patch using robust percentile range.

    Parameters
    ----------
    patch : np.ndarray
        2D image patch.
    upper_percentile : float, optional
        Upper percentile for normalization range.
    lower_percentile : float, optional
        Lower percentile for normalization range.

    Returns
    -------
    np.ndarray
        Patch normalized to ~[0, 1].
    """
    upper_percentile_val = np.percentile(patch, upper_percentile)
    lower_percentile_val = np.percentile(patch, lower_percentile)
    robust_range = np.abs(upper_percentile_val - lower_percentile_val)
    if upper_percentile_val == lower_percentile_val:
        patch = (patch - patch.min()) / (patch.ptp() + 1e-9)
    else:
        patch = (patch - patch.min()) / (robust_range + 1e-9)
    return patch


def split_into_patches_exhaustive(
        scan: np.ndarray,
        pixel_spacing: float,
        patch_edge_len: Union[int, float] = 26,
        overlap_param: float = 0.4,
        patch_size: Tuple[int, int] = (224, 224),
        using_resnet: bool = True,
        ) -> Tuple[torch.Tensor, List[List[Dict]]]:
    """
    Takes in a 3d scan volume and splits it into patches, resizes them then
    normalises them to be passed to the detection network.

    Parameters
    ----------
    scan : np.ndarray
        3d scan volume to be split into patches
    pixel_spacing : float
        The pixel spacing of the scan in mm. Used to calculate patch edge len
    patch_edge_len : Union[int,float]
        The edge length of the patches in cm. Default is 26cm, but can be
        adjusted to be slightly smaller or larger without performance degradation.
    overlap_param : float
        The amount of overlap to create between patches.
    remove_black_space : bool
        Whether to remove black space from the scan.
    patch_size : Tuple[int,int]
        The size of the patches to be created.
    using_resnet : bool
        Changes patch normalization to match the method used to train the VFR ResNet.

    Returns
    -------
    patches : torch.Tensor
        A tensor of patches of shape (num_slices, num_patches, patch_size[0], patch_size[1])
    transform_info_dicts : List[List[Dict]]
        Lists of dictionaries containing the transform information for each patch in each slice.
        Has keys 'x1','x2','y1','y2' indication the position of the patch within the slice
    """

    h, w, d = scan.shape

    if  isinstance(pixel_spacing, (list, tuple, np.ndarray)):
        pixel_spacing = float(pixel_spacing[0])

    if pixel_spacing != -1:
        patch_edge_len = int(patch_edge_len * 10 / pixel_spacing)

    if patch_edge_len > min(scan.shape[0], scan.shape[1]):
        patch_edge_len = min(scan.shape[0:2]) - 1

    # effective_edge_len = how far patches should be spaced from each other
    effective_patch_edge_len = int(patch_edge_len * (1 - overlap_param))

    # work out tiling for scan
    num_patches_across = (w // effective_patch_edge_len) + 1
    num_patches_down = (h // effective_patch_edge_len) + 1
    # total number of patches in each slice
    num_patches = num_patches_down * num_patches_across

    transform_info_dicts = [[None] * num_patches for slice_no in range(d)]
    patches = [[None] * num_patches for slice_no in range(d)]

    for slice_idx in range(d):
        for i in range(num_patches_across):
            x1 = i * effective_patch_edge_len
            x2 = x1 + patch_edge_len
            if x2 >= w:
                x2 = -1
                x1 = -(patch_edge_len)
            for j in range(num_patches_down):
                y1 = j * effective_patch_edge_len
                y2 = y1 + patch_edge_len
                if y2 >= h:
                    y2 = -1
                    y1 = -(patch_edge_len)
                this_patch = np.array(scan[y1:y2, x1:x2, slice_idx])
                resized_patch = cv2.resize(
                    this_patch, patch_size, interpolation=cv2.INTER_CUBIC
                )
                resized_patch[resized_patch < this_patch.min()] = this_patch.min()
                resized_patch[resized_patch > this_patch.max()] = this_patch.max()

                if not using_resnet:
                    patches[slice_idx][i * num_patches_down + j] = 0.5 * torch.Tensor(
                        (resized_patch - np.min(resized_patch))
                        / (np.ptp(resized_patch))
                    )
                else:
                    patches[slice_idx][i * num_patches_down + j] = torch.Tensor(
                        normalize_patch(resized_patch)
                    )
                transform_info_dicts[slice_idx][i * num_patches_down + j] = {
                    "x1": x1,
                    "x2": x2,
                    "y1": y1,
                    "y2": y2,
                }

    return patches, transform_info_dicts


def split_into_patches_exhaustive_spacing(
        scan: np.ndarray,
        pixel_spacing: Union[List[float], float],
        patch_edge_len: Union[int, float] = 26,
        overlap_param: float = 0.4,
        patch_size: Tuple[int, int] = (224, 224),
        using_resnet: bool = True,
) -> Tuple[List[List[torch.Tensor]], List[List[Dict]]]:
    """
    Exhaustively split 3D scan volume into resized, normalized patches for detection.
    Ensures consistency with split_into_patches_exhaustive when px=py.

    Parameters
    ----------
    scan : np.ndarray
        3D scan volume (H, W, D).
    pixel_spacing : Union[List[float], float]
        List [row_spacing, col_spacing] in mm, or scalar.
    patch_edge_len : float, optional
        Patch edge in cm (converted to mm/pixels).
    overlap_param : float, optional
        Overlap fraction between patches.
    patch_size : tuple of int, optional
        Output patch size for resizing (height, width).
    using_resnet : bool, optional
        If True, use robust normalization suited for ResNet training.

    Returns
    -------
    patches : List[List[torch.Tensor]]
        List per slice, each is a list of normalized tensor patches.
    transform_info_dicts : List[List[Dict]]
        Patch spatial origin info per slice.
    """
    h, w, d = scan.shape
    # Interpret pixel_spacing input robustly and support separate row/col spacings
    if len(pixel_spacing) == 2:
        px = float(pixel_spacing[0])  # Row spacing
        py = float(pixel_spacing[-1])  # Column spacing
    elif isinstance(pixel_spacing, (list, tuple, np.ndarray)):
        px = py = float(pixel_spacing[0])
    else:
        # Scalar case
        px = py = float(pixel_spacing)

    # If spacing is sentinel -1 (as used elsewhere), fall back to pixel units
    if px != -1 and py != -1:
        # patch_edge_len is given in cm -> convert to mm then to pixels per axis
        patch_edge_len_mm = patch_edge_len * 10.0
        patch_edge_h = int(np.round(patch_edge_len_mm / px))  # height in pixels
        patch_edge_w = int(np.round(patch_edge_len_mm / py))  # width in pixels
    else:
        # treat provided patch_edge_len as pixels if spacing not provided
        patch_edge_h = patch_edge_w = int(patch_edge_len)

    # Ensure patch edge does not exceed image dims
    patch_edge_h = min(patch_edge_h, h - 1)
    patch_edge_w = min(patch_edge_w, w - 1)

    # Effective stride along each axis
    stride_h = max(1, int(patch_edge_h * (1 - overlap_param)))
    stride_w = max(1, int(patch_edge_w * (1 - overlap_param)))

    # Compute number of patches along each axis
    num_patches_across = (w // stride_w) + 1
    num_patches_down = (h // stride_h) + 1
    num_patches = num_patches_down * num_patches_across

    transform_info_dicts = [[None] * num_patches for _ in range(d)]
    patches = [[None] * num_patches for _ in range(d)]

    for slice_idx in range(d):
        for i in range(num_patches_across):
            x1 = i * stride_w
            x2 = x1 + patch_edge_w
            if x2 >= w:
                x2 = -1
                x1 = -patch_edge_w
            for j in range(num_patches_down):
                y1 = j * stride_h
                y2 = y1 + patch_edge_h
                if y2 >= h:
                    y2 = -1
                    y1 = -patch_edge_h

                this_patch = np.array(scan[y1:y2, x1:x2, slice_idx])

                # In rare cases patch may be empty due to shape issues; skip gracefully
                if this_patch.size == 0:
                    resized_patch = np.zeros(patch_size, dtype=scan.dtype)
                else:
                    resized_patch = cv2.resize(
                        this_patch, patch_size, interpolation=cv2.INTER_CUBIC
                    )
                    # Clip to original patch value range
                    resized_patch[resized_patch < this_patch.min()] = this_patch.min()
                    resized_patch[resized_patch > this_patch.max()] = this_patch.max()

                # Normalize and convert to torch tensor
                if not using_resnet:
                    # protect against zero dynamic range
                    ptp = np.ptp(resized_patch)
                    if ptp == 0:
                        normed = np.zeros_like(resized_patch, dtype=np.float32)
                    else:
                        normed = 0.5 * ((resized_patch - resized_patch.min()) / (ptp))
                    patches[slice_idx][i * num_patches_down + j] = torch.tensor(
                        normed, dtype=torch.float32
                    )
                else:
                    patches[slice_idx][i * num_patches_down + j] = torch.tensor(
                        normalize_patch(resized_patch), dtype=torch.float32
                    )

                transform_info_dicts[slice_idx][i * num_patches_down + j] = {
                    "x1": x1,
                    "x2": x2,
                    "y1": y1,
                    "y2": y2,
                    "patch_edge_h": patch_edge_h,
                    "patch_edge_w": patch_edge_w,
                    "stride_h": stride_h,
                    "stride_w": stride_w,
                }

    return patches, transform_info_dicts


import os
import warnings

import requests
from tqdm import tqdm


def download(url: str, fname: str, verbose: bool = True) -> None:
    '''
    Download a file from a url and save it to a file.

    Parameters
    ----------
    url : str
        The url to download the file from.
    fname : str
        The path to save the file to.
    verbose : bool, optional
        Whether to print progress information. The default is True.
    '''
    resp = requests.get(url, stream=True)
    total = int(resp.headers.get('content-length', 0))
    # Can also replace 'file' with a io.BytesIO object
    if verbose:
        print(f"Downloading {fname} from {url}...")
        with open(fname, 'wb') as file, tqdm(
            desc=fname,
            total=total,
            unit='iB',
            unit_scale=True,
            unit_divisor=1024,
        ) as bar:
            for data in resp.iter_content(chunk_size=1024):
                size = file.write(data)
                bar.update(size)
    else:
        with open(fname, 'wb') as file:
            for data in resp.iter_content(chunk_size=1024):
                file.write(data)
    return



def download_weights(weights_path: str, weight_urls_dict: dict, force: bool = False, verbose: bool = True) -> None:
    '''
    Download weights from the urls in weight_urls_dict to the weights_path.

    Parameters
    ----------
    weights_path : str
        The path to all weights to.
    weight_urls_dict : dict
        The dictionary of urls to download weights from. The key are the filepaths of the weights, and the values are the urls to download the weights from.
    force : bool, optional
        Whether to force download weights even if the weights already exist. The default is False.
    verbose : bool, optional
        Whether to print progress information. The default is True.
    '''
    print('-' * 80)
    print('Downloading weights...')
    print("Note: By downloading these weights and/or adapting SpineNet's source code, you agree to the licence which can be found here: https://github.com/rwindsor1/SpineNet/blob/main/LICENCE.md")
    print('SpineNet is not a diagnostics tool nor a medical device. It should only be used for research.')
    print('-' * 80)
    for path, url in weight_urls_dict.items():
        fname = os.path.join(weights_path, path)
        if not os.path.exists(fname) or force:
            download(url, fname, verbose=verbose)
        else:
            warnings.warn('{} already exists. Skipping download.'.format(fname))
            return



import glob
import os
import torch
import numpy as np


def classify_ivd(classificationNet, ivds, device):
    # Classify
    pred_pf = torch.Tensor().to(device).long()
    pred_nar = torch.Tensor().to(device).long()
    pred_ccs = torch.Tensor().to(device).long()
    pred_spon = torch.Tensor().to(device).long()
    pred_ued = torch.Tensor().to(device).long()
    pred_led = torch.Tensor().to(device).long()
    pred_umc = torch.Tensor().to(device).long()
    pred_lmc = torch.Tensor().to(device).long()
    with torch.no_grad():
        for image in ivds:
            # Augmentations
            num_rows, num_cols, num_slices = image.shape
            max_cols = num_cols - 48
            min_cols = 48
            max_rows = num_rows - 40
            min_rows = 40

            cum_image = arr = np.zeros((1, 1, 9, 112, 224), np.float32)
            for col_s in range(-16, 16 + 1, 16):
                for row_s in range(-16, 16 + 1, 16):
                    # 112 x 224 x 9 Slices
                    image1_1 = image[
                        min_rows + row_s : max_rows + row_s,
                        min_cols + col_s : max_cols + col_s,
                        2:11,
                    ]
                    image2_1 = image[
                        min_rows + row_s : max_rows + row_s,
                        min_cols + col_s : max_cols + col_s,
                        3:12,
                    ]
                    image3_1 = image[
                        min_rows + row_s : max_rows + row_s,
                        min_cols + col_s : max_cols + col_s,
                        4:13,
                    ]
                    image1_2 = np.flip(image1_1, axis=2).copy()
                    image2_2 = np.flip(image2_1, axis=2).copy()
                    image3_2 = np.flip(image3_1, axis=2).copy()

                    # Ready the images
                    image1_1 = (
                        np.transpose(image1_1, (2, 0, 1))[None, None, :, :, :] - 0.2386
                    )  # Channel mean = 0.2386
                    image2_1 = (
                        np.transpose(image2_1, (2, 0, 1))[None, None, :, :, :] - 0.2386
                    )  # Channel mean = 0.2386
                    image3_1 = (
                        np.transpose(image3_1, (2, 0, 1))[None, None, :, :, :] - 0.2386
                    )  # Channel mean = 0.2386
                    image1_2 = (
                        np.transpose(image1_2, (2, 0, 1))[None, None, :, :, :] - 0.2386
                    )  # Channel mean = 0.2386
                    image2_2 = (
                        np.transpose(image2_2, (2, 0, 1))[None, None, :, :, :] - 0.2386
                    )  # Channel mean = 0.2386
                    image3_2 = (
                        np.transpose(image3_2, (2, 0, 1))[None, None, :, :, :] - 0.2386
                    )  # Channel mean = 0.2386
                    cum_image = np.concatenate(
                        (
                            cum_image,
                            image1_1,
                            image2_1,
                            image3_1,
                            image1_2,
                            image2_2,
                            image3_2,
                        ),
                        0,
                    )

            image = np.delete(cum_image, 0, axis=0)
            image = torch.tensor(image).to(device)

            # forward
            net_output = classificationNet(image.float())

            # accumulate prediction and labels
            _, predicted_pf = torch.sum(net_output[0], 0).max(0)
            _, predicted_nar = torch.sum(net_output[1], 0).max(0)
            _, predicted_ccs = torch.sum(net_output[2], 0).max(0)
            _, predicted_spon = torch.sum(net_output[3], 0).max(0)
            _, predicted_ued = torch.sum(net_output[4], 0).max(0)
            _, predicted_led = torch.sum(net_output[5], 0).max(0)
            _, predicted_umc = torch.sum(net_output[6], 0).max(0)
            _, predicted_lmc = torch.sum(net_output[7], 0).max(0)
            pred_pf = torch.cat((pred_pf, predicted_pf[None]), 0)
            pred_nar = torch.cat((pred_nar, predicted_nar[None]), 0)
            pred_ccs = torch.cat((pred_ccs, predicted_ccs[None]), 0)
            pred_spon = torch.cat((pred_spon, predicted_spon[None]), 0)
            pred_ued = torch.cat((pred_ued, predicted_ued[None]), 0)
            pred_led = torch.cat((pred_led, predicted_led[None]), 0)
            pred_umc = torch.cat((pred_umc, predicted_umc[None]), 0)
            pred_lmc = torch.cat((pred_lmc, predicted_lmc[None]), 0)

    pred_pf = pred_pf.cpu().numpy()
    pred_nar = pred_nar.cpu().numpy()
    pred_ccs = pred_ccs.cpu().numpy()
    pred_spon = pred_spon.cpu().numpy()
    pred_ued = pred_ued.cpu().numpy()
    pred_led = pred_led.cpu().numpy()
    pred_umc = pred_umc.cpu().numpy()
    pred_lmc = pred_lmc.cpu().numpy()

    gradings = {}
    gradings["Pfirrmann"] = pred_pf
    gradings["Narrowing"] = pred_nar
    gradings["CentralCanalStenosis"] = pred_ccs
    gradings["Spondylolisthesis"] = pred_spon
    gradings["UpperEndplateDefect"] = pred_ued
    gradings["LowerEndplateDefect"] = pred_led
    gradings["UpperMarrow"] = pred_umc
    gradings["LowerMarrow"] = pred_lmc
    return gradings


def format_volume_for_classification_net(ivd_volume):
    num_rows, num_cols, num_slices = ivd_volume.shape
    max_cols = num_cols - 48
    min_cols = 48
    max_rows = num_rows - 40
    min_rows = 40
    image = ivd_volume[min_rows:max_rows, min_cols:max_cols, 3:12]
    image = torch.tensor(np.transpose(image, (2, 0, 1))[None, None, :, :, :])
    return image


def classify_ivd_v2_resnet(classificationNet, ivds, device):
    pred_pf = torch.Tensor().to(device).long()
    pred_nar = torch.Tensor().to(device).long()
    pred_ccs = torch.Tensor().to(device).long()
    pred_spon = torch.Tensor().to(device).long()
    pred_ued = torch.Tensor().to(device).long()
    pred_led = torch.Tensor().to(device).long()
    pred_umc = torch.Tensor().to(device).long()
    pred_lmc = torch.Tensor().to(device).long()
    pred_fsl = torch.Tensor().to(device).long()
    pred_fsr = torch.Tensor().to(device).long()
    pred_hrn = torch.Tensor().to(device).long()
    for ivd in ivds:
        with torch.no_grad():
            # Augmentations
            image = torch.tensor(ivd)[None, None, :, :, :].float().to(device)
            # forward
            net_output = classificationNet(image)
            # accumulate prediction and labels
            _, predicted_pf = net_output[0].squeeze().max(0)
            _, predicted_nar = net_output[1].squeeze().max(0)
            _, predicted_ccs = net_output[2].squeeze().max(0)
            _, predicted_spon = net_output[3].squeeze().max(0)
            _, predicted_ued = net_output[4].squeeze().max(0)
            _, predicted_led = net_output[5].squeeze().max(0)
            _, predicted_umc = net_output[6].squeeze().max(0)
            _, predicted_lmc = net_output[7].squeeze().max(0)
            _, predicted_fsl = net_output[8].squeeze().max(0)
            _, predicted_fsr = net_output[9].squeeze().max(0)
            _, predicted_hrn = net_output[10].squeeze().max(0)
            pred_pf = torch.cat((pred_pf, predicted_pf[None]), 0)
            pred_nar = torch.cat((pred_nar, predicted_nar[None]), 0)
            pred_ccs = torch.cat((pred_ccs, predicted_ccs[None]), 0)
            pred_spon = torch.cat((pred_spon, predicted_spon[None]), 0)
            pred_ued = torch.cat((pred_ued, predicted_ued[None]), 0)
            pred_led = torch.cat((pred_led, predicted_led[None]), 0)
            pred_umc = torch.cat((pred_umc, predicted_umc[None]), 0)
            pred_lmc = torch.cat((pred_lmc, predicted_lmc[None]), 0)
            pred_fsl = torch.cat((pred_fsl, predicted_fsl[None]), 0)
            pred_fsr = torch.cat((pred_fsr, predicted_fsr[None]), 0)
            pred_hrn = torch.cat((pred_hrn, predicted_hrn[None]), 0)
            torch.cuda.empty_cache()

    pred_pf = pred_pf.cpu().numpy()
    pred_nar = pred_nar.cpu().numpy()
    pred_ccs = pred_ccs.cpu().numpy()
    pred_spon = pred_spon.cpu().numpy()
    pred_ued = pred_ued.cpu().numpy()
    pred_led = pred_led.cpu().numpy()
    pred_umc = pred_umc.cpu().numpy()
    pred_lmc = pred_lmc.cpu().numpy()
    pred_fsl = pred_fsl.cpu().numpy()
    pred_fsr = pred_fsr.cpu().numpy()
    pred_hrn = pred_hrn.cpu().numpy()

    gradings = {}
    gradings["Pfirrmann"] = pred_pf
    gradings["Narrowing"] = pred_nar
    gradings["CentralCanalStenosis"] = pred_ccs
    gradings["Spondylolisthesis"] = pred_spon
    gradings["UpperEndplateDefect"] = pred_ued
    gradings["LowerEndplateDefect"] = pred_led
    gradings["UpperMarrow"] = pred_umc
    gradings["LowerMarrow"] = pred_lmc
    gradings["ForaminalStenosisLeft"] = pred_fsl
    gradings["ForaminalStenosisRight"] = pred_fsr
    gradings["Herniation"] = pred_hrn
    return gradings


def classify_ivd_no_aug(classificationNet, ivds, device):
    # Classify
    pred_pf = torch.Tensor().to(device).long()
    pred_nar = torch.Tensor().to(device).long()
    pred_ccs = torch.Tensor().to(device).long()
    pred_spon = torch.Tensor().to(device).long()
    pred_ued = torch.Tensor().to(device).long()
    pred_led = torch.Tensor().to(device).long()
    pred_umc = torch.Tensor().to(device).long()
    pred_lmc = torch.Tensor().to(device).long()
    with torch.no_grad():
        for image in ivds:
            # Augmentations
            num_rows, num_cols, num_slices = image.shape
            min_cols = 47
            max_cols = 271
            min_rows = 39
            max_rows = 151

            image = image[min_rows:max_rows, min_cols:max_cols, 3:12] - 0.2386
            image = torch.tensor(
                np.transpose(image, (2, 0, 1))[None, None, :, :, :]
            ).to(device)

            # forward
            net_output = classificationNet(image.float())

            # accumulate prediction and labels
            _, predicted_pf = net_output[0].squeeze().max(0)
            _, predicted_nar = net_output[1].squeeze().max(0)
            _, predicted_ccs = net_output[2].squeeze().max(0)
            _, predicted_spon = net_output[3].squeeze().max(0)
            _, predicted_ued = net_output[4].squeeze().max(0)
            _, predicted_led = net_output[5].squeeze().max(0)
            _, predicted_umc = net_output[6].squeeze().max(0)
            _, predicted_lmc = net_output[7].squeeze().max(0)
            pred_pf = torch.cat((pred_pf, predicted_pf[None]), 0)
            pred_nar = torch.cat((pred_nar, predicted_nar[None]), 0)
            pred_ccs = torch.cat((pred_ccs, predicted_ccs[None]), 0)
            pred_spon = torch.cat((pred_spon, predicted_spon[None]), 0)
            pred_ued = torch.cat((pred_ued, predicted_ued[None]), 0)
            pred_led = torch.cat((pred_led, predicted_led[None]), 0)
            pred_umc = torch.cat((pred_umc, predicted_umc[None]), 0)
            pred_lmc = torch.cat((pred_lmc, predicted_lmc[None]), 0)

    pred_pf = pred_pf.cpu().numpy()
    pred_nar = pred_nar.cpu().numpy()
    pred_ccs = pred_ccs.cpu().numpy()
    pred_spon = pred_spon.cpu().numpy()
    pred_ued = pred_ued.cpu().numpy()
    pred_led = pred_led.cpu().numpy()
    pred_umc = pred_umc.cpu().numpy()
    pred_lmc = pred_lmc.cpu().numpy()

    gradings = {}
    gradings["Pfirrmann"] = pred_pf
    gradings["Narrowing"] = pred_nar
    gradings["CentralCanalStenosis"] = pred_ccs
    gradings["Spondylolisthesis"] = pred_spon
    gradings["UpperEndplateDefect"] = pred_ued
    gradings["LowerEndplateDefect"] = pred_led
    gradings["UpperMarrow"] = pred_umc
    gradings["LowerMarrow"] = pred_lmc
    return gradings


def classify_ivd_no_aug_spinenetV1(classificationNet, ivds, device):
    # Classify
    pred_pf = torch.Tensor().to(device).long()
    pred_nar = torch.Tensor().to(device).long()
    pred_ccs = torch.Tensor().to(device).long()
    pred_spon = torch.Tensor().to(device).long()
    pred_ued = torch.Tensor().to(device).long()
    pred_led = torch.Tensor().to(device).long()
    pred_umc = torch.Tensor().to(device).long()
    pred_lmc = torch.Tensor().to(device).long()
    with torch.no_grad():
        for image in ivds:
            # Augmentations
            num_rows, num_cols, num_slices = image.shape
            min_cols = 47
            max_cols = 271
            min_rows = 39
            max_rows = 151

            # image = image[min_rows:max_rows,min_cols:max_cols,3:12] - 0.2386
            # SpineNet V1 normalization
            image = image[min_rows:max_rows, min_cols:max_cols, 3:12]
            image = torch.tensor(
                np.transpose(image, (2, 0, 1))[None, None, :, :, :]
            ).to(device)

            # forward
            net_output = classificationNet(image.float())

            # accumulate prediction and labels
            _, predicted_pf = net_output[0].squeeze().max(0)
            _, predicted_nar = net_output[1].squeeze().max(0)
            _, predicted_ccs = net_output[2].squeeze().max(0)
            _, predicted_spon = net_output[3].squeeze().max(0)
            _, predicted_ued = net_output[4].squeeze().max(0)
            _, predicted_led = net_output[5].squeeze().max(0)
            _, predicted_umc = net_output[6].squeeze().max(0)
            _, predicted_lmc = net_output[7].squeeze().max(0)
            pred_pf = torch.cat((pred_pf, predicted_pf[None]), 0)
            pred_nar = torch.cat((pred_nar, predicted_nar[None]), 0)
            pred_ccs = torch.cat((pred_ccs, predicted_ccs[None]), 0)
            pred_spon = torch.cat((pred_spon, predicted_spon[None]), 0)
            pred_ued = torch.cat((pred_ued, predicted_ued[None]), 0)
            pred_led = torch.cat((pred_led, predicted_led[None]), 0)
            pred_umc = torch.cat((pred_umc, predicted_umc[None]), 0)
            pred_lmc = torch.cat((pred_lmc, predicted_lmc[None]), 0)

    pred_pf = pred_pf.cpu().numpy()
    pred_nar = pred_nar.cpu().numpy()
    pred_ccs = pred_ccs.cpu().numpy()
    pred_spon = pred_spon.cpu().numpy()
    pred_ued = pred_ued.cpu().numpy()
    pred_led = pred_led.cpu().numpy()
    pred_umc = pred_umc.cpu().numpy()
    pred_lmc = pred_lmc.cpu().numpy()

    gradings = {}
    gradings["Pfirrmann"] = pred_pf
    gradings["Narrowing"] = pred_nar
    gradings["CentralCanalStenosis"] = pred_ccs
    gradings["Spondylolisthesis"] = pred_spon
    gradings["UpperEndplateDefect"] = pred_ued
    gradings["LowerEndplateDefect"] = pred_led
    gradings["UpperMarrow"] = pred_umc
    gradings["LowerMarrow"] = pred_lmc
    return gradings


def classify_ivd_spinenetV1(classificationNet, ivds, device):
    # Classify
    pred_pf = torch.Tensor().to(device).long()
    pred_nar = torch.Tensor().to(device).long()
    pred_ccs = torch.Tensor().to(device).long()
    pred_spon = torch.Tensor().to(device).long()
    pred_ued = torch.Tensor().to(device).long()
    pred_led = torch.Tensor().to(device).long()
    pred_umc = torch.Tensor().to(device).long()
    pred_lmc = torch.Tensor().to(device).long()
    with torch.no_grad():
        for image in ivds:
            # Augmentations
            num_rows, num_cols, num_slices = image.shape
            max_cols = num_cols - 48
            min_cols = 48
            max_rows = num_rows - 40
            min_rows = 40

            cum_image = arr = np.zeros((1, 1, 9, 112, 224), np.float32)
            for col_s in range(-16, 16 + 1, 16):
                for row_s in range(-16, 16 + 1, 16):
                    # 112 x 224 x 9 Slices
                    image1_1 = image[
                        min_rows + row_s : max_rows + row_s,
                        min_cols + col_s : max_cols + col_s,
                        2:11,
                    ]
                    image2_1 = image[
                        min_rows + row_s : max_rows + row_s,
                        min_cols + col_s : max_cols + col_s,
                        3:12,
                    ]
                    image3_1 = image[
                        min_rows + row_s : max_rows + row_s,
                        min_cols + col_s : max_cols + col_s,
                        4:13,
                    ]
                    image1_2 = np.flip(image1_1, axis=2).copy()
                    image2_2 = np.flip(image2_1, axis=2).copy()
                    image3_2 = np.flip(image3_1, axis=2).copy()

                    # Ready the images
                    image1_1 = np.transpose(image1_1, (2, 0, 1))[
                        None, None, :, :, :
                    ]  # Channel mean = 0.2386
                    image2_1 = np.transpose(image2_1, (2, 0, 1))[
                        None, None, :, :, :
                    ]  # Channel mean = 0.2386
                    image3_1 = np.transpose(image3_1, (2, 0, 1))[
                        None, None, :, :, :
                    ]  # Channel mean = 0.2386
                    image1_2 = np.transpose(image1_2, (2, 0, 1))[
                        None, None, :, :, :
                    ]  # Channel mean = 0.2386
                    image2_2 = np.transpose(image2_2, (2, 0, 1))[
                        None, None, :, :, :
                    ]  # Channel mean = 0.2386
                    image3_2 = np.transpose(image3_2, (2, 0, 1))[
                        None, None, :, :, :
                    ]  # Channel mean = 0.2386
                    cum_image = np.concatenate(
                        (
                            cum_image,
                            image1_1,
                            image2_1,
                            image3_1,
                            image1_2,
                            image2_2,
                            image3_2,
                        ),
                        0,
                    )

            image = np.delete(cum_image, 0, axis=0)
            # SpineNet V1 normalization
            image[:, :, 0, :, :] -= 0.1351
            image[:, :, 1, :, :] -= 0.1363
            image[:, :, 2, :, :] -= 0.1371
            image[:, :, 3, :, :] -= 0.1358
            image[:, :, 4, :, :] -= 0.1348
            image[:, :, 5, :, :] -= 0.1365
            image[:, :, 6, :, :] -= 0.1368
            image[:, :, 7, :, :] -= 0.1364
            image[:, :, 8, :, :] -= 0.1357
            image = torch.tensor(image).to(device)

            # forward
            net_output = classificationNet(image.float())

            # accumulate prediction and labels
            _, predicted_pf = torch.sum(net_output[0], 0).max(0)
            _, predicted_nar = torch.sum(net_output[1], 0).max(0)
            _, predicted_ccs = torch.sum(net_output[2], 0).max(0)
            _, predicted_spon = torch.sum(net_output[3], 0).max(0)
            _, predicted_ued = torch.sum(net_output[4], 0).max(0)
            _, predicted_led = torch.sum(net_output[5], 0).max(0)
            _, predicted_umc = torch.sum(net_output[6], 0).max(0)
            _, predicted_lmc = torch.sum(net_output[7], 0).max(0)
            pred_pf = torch.cat((pred_pf, predicted_pf[None]), 0)
            pred_nar = torch.cat((pred_nar, predicted_nar[None]), 0)
            pred_ccs = torch.cat((pred_ccs, predicted_ccs[None]), 0)
            pred_spon = torch.cat((pred_spon, predicted_spon[None]), 0)
            pred_ued = torch.cat((pred_ued, predicted_ued[None]), 0)
            pred_led = torch.cat((pred_led, predicted_led[None]), 0)
            pred_umc = torch.cat((pred_umc, predicted_umc[None]), 0)
            pred_lmc = torch.cat((pred_lmc, predicted_lmc[None]), 0)

    pred_pf = pred_pf.cpu().numpy()
    pred_nar = pred_nar.cpu().numpy()
    pred_ccs = pred_ccs.cpu().numpy()
    pred_spon = pred_spon.cpu().numpy()
    pred_ued = pred_ued.cpu().numpy()
    pred_led = pred_led.cpu().numpy()
    pred_umc = pred_umc.cpu().numpy()
    pred_lmc = pred_lmc.cpu().numpy()

    gradings = {}
    gradings["Pfirrmann"] = pred_pf
    gradings["Narrowing"] = pred_nar
    gradings["CentralCanalStenosis"] = pred_ccs
    gradings["Spondylolisthesis"] = pred_spon
    gradings["UpperEndplateDefect"] = pred_ued
    gradings["LowerEndplateDefect"] = pred_led
    gradings["UpperMarrow"] = pred_umc
    gradings["LowerMarrow"] = pred_lmc
    return gradings



from unittest.mock import patch
import torch
import numpy as np
import torch.nn.functional as F
import scipy.ndimage
from shapely.geometry import Polygon, Point


def validate_coordinates(coords, max_shape):
    """
    Filter out coordinates that are negative or exceed image bounds.

    Parameters
    ----------
    coords : array-like, shape (N, 2)
        List or array of coordinates to validate.
    max_shape : tuple of int
        Maximum allowed shape \(`height`, `width`\) for the coordinates.

    Returns
    -------
    valid_coords : ndarray, shape (M, 2)
        Array of valid coordinates within the image bounds.
    """
    if len(coords) == 0:
        return np.empty((0, 2))

    coords = np.array(coords)
    valid_mask = (
            (coords[:, 0] >= 0) & (coords[:, 0] < max_shape[0]) &
            (coords[:, 1] >= 0) & (coords[:, 1] < max_shape[1])
    )
    return coords[valid_mask]

def make_in_slice_detections(
    detection_net,
    patches,
    transform_info_dicts,
    scan_shape,
    corner_threshold=0.3,
    centroid_threshold=0.5,
    device="cuda:0",
):
    """
    Take normalized, resized patches and pass them through the detection net,
    outputting detections in each slice

    Inputs
    ------
    detection_net : torch.nn.Module
        the VFR model used to detect vertebrae landmarks and corners
    patches: list of list of torch.Tensors
        a S-long list of P equal sized torch.FloatTensors where S indexes the
        slice number and P indexes the patch in that slice. As produced by
        scan_preprocessing.split_into_patches

    Outputs
    -------
    detection_dicts : list of dicts
        dictionaries containing information about detections made in each slice
        by the network.
    """
    # now resize patches
    detection_net.to(device)
    detection_net.eval()
    detection_dicts = []
    patches_dicts = []
    # flatten patches and pass to network
    # patches_flat_list = [patch for slice_list in patches for patch in slice_list]
    # patches_tensor = torch.stack(patches_flat_list, dim=0).unsqueeze(1)

    for slice_idx in range(len(patches)):
        patches_tensor = torch.stack(patches[slice_idx], dim=0)[:, None, :, :]
        flipped_patches_tensor = torch.flip(patches_tensor, [-1])
        net_input = torch.cat([patches_tensor, flipped_patches_tensor], axis=0)
        with torch.no_grad():
            # both_net_output = detection_net(net_input.to(device).float()).cpu()
            net_output = detection_net(patches_tensor.to(device).float()).cpu()

        patches_dicts.append({"patches": patches_tensor.numpy(), "net_output": net_output.numpy(), 
                              "landmark_points": {}, "landmark_arrows": {}})

        all_corners = {"points": {}, "arrows": {}}
        for corner_type in ["rt", "rb", "lb", "lt"]:
            all_corners["points"][corner_type] = []
            all_corners["arrows"][corner_type] = []
            patches_dicts[-1]["landmark_points"][corner_type] = []
            patches_dicts[-1]["landmark_arrows"][corner_type] = []


        scan_centroid_channel = np.zeros(scan_shape[0:2]).astype(float)
        centroid_channel_contributions = np.zeros_like(scan_centroid_channel)
        max_x = transform_info_dicts[slice_idx][0]["x2"]
        min_x = transform_info_dicts[slice_idx][0]["x1"]

        patch_edge_len = np.abs(max_x - min_x)

        # resized the centroid channel to transform it into original frame

        # loop through each patch processed in this slice
        for j in range(len(patches[slice_idx])):
            transform_info = transform_info_dicts[slice_idx][j]
            try:
                # resized the centroid channel to transform it into original frame

                resized_centroid_channels = F.interpolate(
                    net_output[:, 4, :, :].unsqueeze(1),
                    (
                        transform_info["y2"] - transform_info["y1"],
                        transform_info["x2"] - transform_info["x1"],
                    ),
                    mode="bilinear",
                    align_corners=False,
                )
                scan_centroid_channel[
                    transform_info["y1"] : transform_info["y2"],
                    transform_info["x1"] : transform_info["x2"],
                ] = (
                    scan_centroid_channel[
                        transform_info["y1"] : transform_info["y2"],
                        transform_info["x1"] : transform_info["x2"],
                    ]
                    + resized_centroid_channels[j, 0, :, :].numpy()
                )
                centroid_channel_contributions[
                    transform_info["y1"] : transform_info["y2"],
                    transform_info["x1"] : transform_info["x2"],
                ] += 1
            except Exception as E:
                print(
                    resized_centroid_channels.shape,
                    scan_centroid_channel[
                        transform_info["y1"] : transform_info["y2"],
                        transform_info["x1"] : transform_info["x2"],
                    ].shape,
                )
                print(str(E))

            # convert corners into full scan frame
            for corner_idx, corner_type in enumerate(["rt", "rb", "lb", "lt"]):
                # get corners
                points = get_points(
                    net_output[j, corner_idx, :, :], threshold=corner_threshold
                )
                if len(points) == 0:
                    patches_dicts[-1]["landmark_points"][corner_type].append([])
                    patches_dicts[-1]["landmark_arrows"][corner_type].append([])
                    continue
                else:
                    if transform_info["y1"] < 0:
                        transform_info["y1"] = (
                            scan_centroid_channel.shape[0] + transform_info["y1"]
                        )
                    if transform_info["x1"] < 0:
                        transform_info["x1"] = (
                            scan_centroid_channel.shape[1] + transform_info["x1"]
                        )
                    transformed_patch_corners = (
                        points * patch_edge_len / 224
                        + np.array([transform_info["y1"], transform_info["x1"]])
                    )
                    transformed_patch_corners = validate_coordinates(
                        transformed_patch_corners, scan_shape
                    )
                    all_corners["points"][corner_type].append(transformed_patch_corners)
                    arrows = np.zeros_like(points)
                    for idx, point in enumerate(points):
                        arrows[idx, 0] = net_output[
                            j, corner_idx + 9, point[0], point[1]
                        ]
                        arrows[idx, 1] = net_output[
                            j, corner_idx + 5, point[0], point[1]
                        ]
                    # scale arrows to original frame
                    all_corners["arrows"][corner_type].append(
                        arrows * patch_edge_len / 224
                    )
                    patches_dicts[-1]["landmark_points"][corner_type].append(points)
                    patches_dicts[-1]["landmark_arrows"][corner_type].append(arrows)


        # disp_corners contained the displaced corners, i.e. the vector sum
        # of the corner position and its arrow. These displaced corners are used
        # to see which centroid each corner points closest to
        disp_corners = {}
        for i in ["rt", "rb", "lb", "lt"]:
            if len(all_corners["points"][i]) > 0:
                all_corners["points"][i] = np.concatenate(
                    all_corners["points"][i], axis=0
                )
                all_corners["arrows"][i] = np.concatenate(
                    all_corners["arrows"][i], axis=0
                )
                disp_corners[i] = all_corners["points"][i] + all_corners["arrows"][i]
            else:
                disp_corners[i] = []

        # get mean centroid channel from all contributons
        centroid_channel_contributions[centroid_channel_contributions == 0] = 1
        scan_centroid_channel /= centroid_channel_contributions
        # detect the centroids
        centroids = get_points(scan_centroid_channel, threshold=centroid_threshold)
        centroids = validate_coordinates(centroids, scan_shape)

        detection_polys = []
        arrows = all_corners["arrows"]
        corners = all_corners["points"]

        # now loop through each detected centroid and find the closest displaced
        # corner of each type
        for centroid in centroids:
            if (
                len(disp_corners["lb"])
                and len(disp_corners["lt"])
                and len(disp_corners["rb"])
                and len(disp_corners["rt"])
            ):
                lt_dist = np.min(np.linalg.norm(centroid - disp_corners["lt"], axis=1))
                lt_arrow = arrows["lt"][
                    np.argmin(np.linalg.norm(centroid - disp_corners["lt"], axis=1))
                ]
                closest_lt_corner = corners["lt"][
                    np.argmin(np.linalg.norm(centroid - disp_corners["lt"], axis=1))
                ]
                lb_dist = np.min(np.linalg.norm(centroid - disp_corners["lb"], axis=1))
                lb_arrow = arrows["lb"][
                    np.argmin(np.linalg.norm(centroid - disp_corners["lb"], axis=1))
                ]
                closest_lb_corner = corners["lb"][
                    np.argmin(np.linalg.norm(centroid - disp_corners["lb"], axis=1))
                ]
                rt_dist = np.min(np.linalg.norm(centroid - disp_corners["rt"], axis=1))
                rt_arrow = arrows["rt"][
                    np.argmin(np.linalg.norm(centroid - disp_corners["rt"], axis=1))
                ]
                closest_rt_corner = corners["rt"][
                    np.argmin(np.linalg.norm(centroid - disp_corners["rt"], axis=1))
                ]
                rb_dist = np.min(np.linalg.norm(centroid - disp_corners["rb"], axis=1))
                rb_arrow = arrows["rb"][
                    np.argmin(np.linalg.norm(centroid - disp_corners["rb"], axis=1))
                ]
                closest_rb_corner = corners["rb"][
                    np.argmin(np.linalg.norm(centroid - disp_corners["rb"], axis=1))
                ]
                indiv_arrows = [rt_arrow, rb_arrow, lb_arrow, lt_arrow]
                poly = Polygon(
                    [
                        closest_rt_corner,
                        closest_rb_corner,
                        closest_lb_corner,
                        closest_lt_corner,
                    ]
                )
                # threshold arrows, do not allow if distance from centroid is over 0.5* length of arrow
                missing_arrows = arrows_threshold_check(
                    [
                        [rt_dist, rt_arrow],
                        [rb_dist, rb_arrow],
                        [lb_dist, lb_arrow],
                        [lt_dist, lt_arrow],
                    ]
                )
                detection_poly = np.array(
                    [
                        closest_rt_corner,
                        closest_rb_corner,
                        closest_lb_corner,
                        closest_lt_corner,
                    ]
                )
                if sum(missing_arrows) == 0:
                    if poly.is_valid and poly.contains(Point(centroid)):
                        # flip around poly to match form needed for matplotlib plotting
                        detection_polys.append([[i[1], i[0]] for i in detection_poly])
                # # if only one arrow is missing simply assume vertebra is a diamond shape
                elif sum(missing_arrows) == 1:
                    for i, el in enumerate(missing_arrows):
                        if el:
                            new_corner = centroid + indiv_arrows[(i + 2) % 4]
                            # **VALIDATE NEW CORNER COORDINATE*
                            if not (0 <= new_corner[0] < scan_shape[0] and 0 <= new_corner[1] < scan_shape[1]):
                                # Get closest valid coordinate within image bounds
                                new_corner = np.clip(new_corner, [0, 0], np.array(scan_shape[:2]) - 1)
                            detection_poly[i] = new_corner

                    # # flip around poly to match form needed for matplotlib plotting
                    detection_polys.append([[i[1], i[0]] for i in detection_poly])

        # loop through each detection poly and remove those where the same corner
        # is shared between two polygons
        # else: print(slice_idx, len(patches)//2)
        detection_polys = remove_polys_sharing_corners(detection_polys, all_corners)

        all_corners["centroid_heatmap"] = scan_centroid_channel
        all_corners["centroids"] = centroids
        all_corners["detection_polys"] = detection_polys
        all_corners["vert_index"] = [None] * len(detection_polys)
        detection_dicts.append(all_corners)

    return detection_dicts, patches_dicts


def remove_polys_sharing_corners(detection_polys, all_corners):
    """
    Takes as input all the polygons detected in the slice and where two polygons
    share the same corner, remove one of them. Of the two overlapping polygons,
    the one with the biggest difference between its largest internal angle and
    its smallest is removed.

    Inputs
    ------
    detection_polys : list of 4x2 numpy float arrays
        all the polygons detected in the slice so far
    all_corners : dictionary
        information about the arrows for each polygon vertex.

    Outputs
    -------
    detection_polys : list of 4x2 lists.
        A new polygon list with spurious detections removed.
    """
    # stores indicies of the polys to be removed
    polys_to_be_removed = []
    for detection_poly_idx, detection_poly in enumerate(detection_polys):
        for other_detection_poly_idx, other_detection_poly in enumerate(
            detection_polys
        ):
            # skip over overlaps with its self
            if detection_poly != other_detection_poly:
                # vectors between all the corners
                diff_arr = np.array(detection_poly) - np.array(other_detection_poly)
                if np.min(np.linalg.norm(diff_arr, axis=-1)) < 2:
                    # should go here if the detection polys have the same corner
                    detection_poly_internal_angles = get_internal_angles(detection_poly)
                    other_detection_poly_internal_angles = get_internal_angles(
                        other_detection_poly
                    )

                    if np.ptp(detection_poly_internal_angles) > np.ptp(
                        other_detection_poly_internal_angles
                    ):
                        polys_to_be_removed.append(detection_poly_idx)
                    else:
                        polys_to_be_removed.append(other_detection_poly_idx)

    polys_to_be_removed = sorted(list(set(polys_to_be_removed)), reverse=True)

    for poly_to_be_removed in polys_to_be_removed:
        detection_polys.pop(poly_to_be_removed)

    return detection_polys


def get_internal_angles(poly):
    """
    Calculates the internal angles for a quadrilateral shaped as a 4x2 numpy list

    Inputs
    ------
    poly : 4x2 numpy array
        the polygon to calculate the internal angles of
    Outputs
    -------
    poly_internal_angles : list
        a list of internal angles for the polygon
    """
    poly_internal_angles = []
    poly = np.array(poly)
    for i in range(4):
        v1 = poly[i - 1, :] - poly[i, :]
        v2 = poly[(i + 1) % 4, :] - poly[i, :]
        v1 /= np.linalg.norm(v1)
        v2 /= np.linalg.norm(v2)
        poly_internal_angles.append(np.arccos(np.dot(v1, v2)))

    return poly_internal_angles


def arrows_threshold_check(arrows_arr):
    """
    find corners without an arrow pointing to the centroid within a certain threshold distance
    """
    # TODO: Complete docstring
    # arrows arr should be a list of 2-entry lists containing the distance form the centroid and the arrow in cartesian coords
    missing_arrows = [0, 0, 0, 0]  # binary mask to show missing corner
    for idx, el in enumerate(arrows_arr):
        if el[0] > 0.5 * np.linalg.norm(el[1]) or el[0] > 20:
            missing_arrows[idx] = 1
    return missing_arrows


def get_points(image, threshold=0.5):
    """
    finds points from gaussian response maps

    this is done by thresholding the response map to get a binary map and then
    treating each interconnected volume as a detection. The point of maximum
    response in each interconnected volume becomes the location of the landmark

    Inputs
    ------
    image : np.array-like
        HxW image containing gaussian response map for landmarks
    threshold: float
        the value at which to threshold the response map.

    Outputs
    -------
    points : a Nx2 array containing the locations of each point detected
    """

    image = np.array(image)
    mask = image > threshold
    segmentation, no_labels = scipy.ndimage.label(mask)
    points = []
    for label in range(no_labels):
        point = np.unravel_index(
            np.argmax(image * (segmentation == (label + 1))), image.shape
        )
        points.append(point)

    return np.array(points)



import torchvision
import os, glob
import torch
import torch.nn as nn

resnet = torchvision.models.resnet.resnet50(pretrained=False)


class ConvBlock(nn.Module):
    """
    Helper module that consists of a Conv -> BN -> ReLU
    """

    def __init__(
        self,
        in_channels,
        out_channels,
        padding=1,
        kernel_size=3,
        stride=1,
        with_nonlinearity=True,
    ):
        super().__init__()
        self.conv = nn.Conv2d(
            in_channels,
            out_channels,
            padding=padding,
            kernel_size=kernel_size,
            stride=stride,
        )
        self.bn = nn.BatchNorm2d(out_channels)
        self.relu = nn.ReLU()
        self.with_nonlinearity = with_nonlinearity

    def forward(self, x):
        x = self.conv(x)
        x = self.bn(x)
        if self.with_nonlinearity:
            x = self.relu(x)
        return x


class Bridge(nn.Module):
    """
    This is the middle layer of the UNet which just consists of some
    """

    def __init__(self, in_channels, out_channels):
        super().__init__()
        self.bridge = nn.Sequential(
            ConvBlock(in_channels, out_channels), ConvBlock(out_channels, out_channels)
        )

    def forward(self, x):
        return self.bridge(x)


class UpBlockForUNetWithResNet50(nn.Module):
    """
    Up block that encapsulates one up-sampling step which consists of Upsample -> ConvBlock -> ConvBlock
    """

    def __init__(
        self,
        in_channels,
        out_channels,
        up_conv_in_channels=None,
        up_conv_out_channels=None,
        upsampling_method="bilinear",
    ):
        super().__init__()

        if up_conv_in_channels == None:
            up_conv_in_channels = in_channels
        if up_conv_out_channels == None:
            up_conv_out_channels = out_channels

        if upsampling_method == "conv_transpose":
            self.upsample = nn.ConvTranspose2d(
                up_conv_in_channels, up_conv_out_channels, kernel_size=2, stride=2
            )
        elif upsampling_method == "bilinear":
            self.upsample = nn.Sequential(
                nn.Upsample(mode="bilinear", scale_factor=2, align_corners=False),
                nn.Conv2d(
                    up_conv_in_channels, up_conv_out_channels, kernel_size=1, stride=1
                ),
            )
        self.conv_block_1 = ConvBlock(in_channels, out_channels)
        self.conv_block_2 = ConvBlock(out_channels, out_channels)

    def forward(self, up_x, down_x):
        """

        :param up_x: this is the output from the previous up block
        :param down_x: this is the output from the down block
        :return: upsampled feature map
        """
        x = self.upsample(up_x)
        x = torch.cat([x, down_x], 1)
        x = self.conv_block_1(x)
        x = self.conv_block_2(x)
        return x


class VFRResNetDetector(nn.Module):
    DEPTH = 6

    def __init__(self, n_classes=13):
        super().__init__()
        resnet = torchvision.models.resnet.resnet50(pretrained=False)
        down_blocks = []
        up_blocks = []
        self.input_block = nn.Sequential(
            nn.Conv2d(
                1, 64, kernel_size=(7, 7), stride=(2, 2), padding=(3, 3), bias=False
            ),
            nn.BatchNorm2d(
                64, eps=1e-5, momentum=0.1, affine=True, track_running_stats=True
            ),
            nn.ReLU(inplace=True),
        )
        self.input_pool = list(resnet.children())[3]
        for bottleneck in list(resnet.children()):
            if isinstance(bottleneck, nn.Sequential):
                down_blocks.append(bottleneck)
        self.down_blocks = nn.ModuleList(down_blocks)
        self.bridge = Bridge(2048, 2048)
        up_blocks.append(UpBlockForUNetWithResNet50(2048, 1024))
        up_blocks.append(UpBlockForUNetWithResNet50(1024, 512))
        up_blocks.append(UpBlockForUNetWithResNet50(512, 256))
        up_blocks.append(
            UpBlockForUNetWithResNet50(
                in_channels=128 + 64,
                out_channels=128,
                up_conv_in_channels=256,
                up_conv_out_channels=128,
            )
        )
        up_blocks.append(
            UpBlockForUNetWithResNet50(
                in_channels=64 + 1,
                out_channels=64,
                up_conv_in_channels=128,
                up_conv_out_channels=64,
            )
        )

        self.up_blocks = nn.ModuleList(up_blocks)

        self.out = nn.Conv2d(64, n_classes, kernel_size=1, stride=1)

    def forward(self, x, with_output_feature_map=False):
        pre_pools = dict()
        pre_pools[f"layer_0"] = x
        x = self.input_block(x)
        pre_pools[f"layer_1"] = x
        x = self.input_pool(x)

        for i, block in enumerate(self.down_blocks, 2):
            x = block(x)
            if i == (VFRResNetDetector.DEPTH - 1):
                continue
            pre_pools[f"layer_{i}"] = x

        x = self.bridge(x)

        for i, block in enumerate(self.up_blocks, 1):
            key = f"layer_{VFRResNetDetector.DEPTH - 1 - i}"
            x = block(x, pre_pools[key])
        output_feature_map = x
        x = self.out(x)
        del pre_pools
        if with_output_feature_map:
            return x, output_feature_map
        else:
            return x

    def load_weights(self, save_path, verbose=True):
        if os.path.isdir(save_path):
            list_of_pt = [
                x for x in glob.glob(save_path + "/*.pt") if "encrypted" not in x
            ]

            latest_pt = max(list_of_pt, key=os.path.getctime)
            checkpoint = torch.load(latest_pt, map_location="cpu")

            self.load_state_dict(checkpoint["net"])
            best_loss = checkpoint["loss"]
            start_epoch = checkpoint["epoch"] + 1
            if verbose:
                print(f"==> Loading model trained for {start_epoch} epochs...")
        else:
            raise NameError(f"save path {save_path} could not be found")


if __name__ == "__main__":
    print(list(resnet.children())[:3])
    model = VFRResNetDetector()
    print(model.input_block)
    x = torch.rand(1, 1, 512, 512)
    print(x.shape)
    output = model(x)
    print(output.shape)



import torch
import torch.nn as nn
import torch.nn.functional as F
import os, glob


class ContextModel(nn.Module):
    def __init__(self):
        super(ContextModel, self).__init__()
        # down
        self.conv1 = nn.Conv2d(1, 16, (5, 3), dilation=(2, 1), padding=(4, 1))
        self.bnrm1 = nn.BatchNorm2d(16)
        self.conv2 = nn.Conv2d(16, 32, (5, 3), dilation=(2, 1), padding=(4, 1))
        self.bnrm2 = nn.BatchNorm2d(32)
        self.conv3 = nn.Conv2d(32, 64, (5, 3), dilation=(2, 1), padding=(4, 1))
        self.bnrm3 = nn.BatchNorm2d(64)
        self.conv4 = nn.Conv2d(64, 128, (5, 3), dilation=(2, 1), padding=(4, 1))
        self.bnrm4 = nn.BatchNorm2d(128)

        # mid conv
        self.conv_mid = nn.Conv2d(128, 128, (3, 3), dilation=(1, 1), padding=(1, 1))
        self.bnrm_mid = nn.BatchNorm2d(128)
        # up
        self.conv5 = nn.Conv2d(128, 64, (5, 3), dilation=(2, 1), padding=(4, 1))
        self.bnrm5 = nn.BatchNorm2d(64)
        self.conv6 = nn.Conv2d(128, 32, (5, 3), dilation=(2, 1), padding=(4, 1))
        self.bnrm6 = nn.BatchNorm2d(32)
        self.conv7 = nn.Conv2d(64, 16, (5, 3), dilation=(2, 1), padding=(4, 1))
        self.bnrm7 = nn.BatchNorm2d(16)
        self.conv8 = nn.Conv2d(16, 1, (5, 3), dilation=(2, 1), padding=(4, 1))

    def forward(self, x):
        # Down
        x = F.max_pool2d(self.bnrm1(F.relu(self.conv1(x))), kernel_size=(2, 1))  # Conv1
        x2 = F.max_pool2d(
            self.bnrm2(F.relu(self.conv2(x))), kernel_size=(2, 1)
        )  # Conv2
        x3 = F.max_pool2d(
            self.bnrm3(F.relu(self.conv3(x2))), kernel_size=(2, 1)
        )  # Conv3
        x = F.max_pool2d(
            self.bnrm4(F.relu(self.conv4(x3))), kernel_size=(2, 1)
        )  # Conv4
        # Mid
        x = self.bnrm_mid(F.relu(self.conv_mid(x)))
        # Up
        x = F.interpolate(x, scale_factor=(2, 1), mode="bilinear", align_corners=False)
        x = self.bnrm5(F.relu(self.conv5(x)))
        x = torch.cat((x, x3), 1)
        x = F.interpolate(x, scale_factor=(2, 1), mode="bilinear", align_corners=False)
        x = self.bnrm6(F.relu(self.conv6(x)))
        x = torch.cat((x, x2), 1)
        x = F.interpolate(x, scale_factor=(2, 1), mode="bilinear", align_corners=False)
        x = self.bnrm7(F.relu(self.conv7(x)))
        x = F.interpolate(x, scale_factor=(2, 1), mode="bilinear", align_corners=False)
        x = self.conv8(x)
        return x

    def load_weights(self, save_path, verbose=True):
        if os.path.isdir(save_path):
            list_of_pt = glob.glob(save_path + "/*.pt")
            latest_pt = max(list_of_pt, key=os.path.getctime)
            checkpoint = torch.load(latest_pt, map_location="cpu")
            self.load_state_dict(checkpoint["net"])
            best_loss = checkpoint["loss"]
            start_epoch = checkpoint["epoch"] + 1
            if verbose:
                print(f"==> Loading model trained for {start_epoch} epochs...")
        else:
            raise NameError(f"save path {save_path} could not be found")
        return



import torch
from shapely.geometry import Polygon as shapelyPolygon
from shapely.geometry import Point as shapelyPoint
import numpy as np
import cv2
import torch.nn.functional as F


def extract_volumes(
    scan,
    vert_dicts,
    extent=1,
    rescale_sagittal=True,
    output_shape=(224, 224, 16),
    resampling_mode="bicubic",
):
    """
    Extracts volumes surrounding each vertebra detected in the scan.

    Inputs
    -----
    scan :  np.array
        A HxWxS array containing the MRI scan. H is height of scan, W is width and S is number of sagittal slices
    vert_dicts :  list of dictionaries
        Each dictionary contains information about the vertebra detections in the scan, as given by the output
        from `utils.group_slice_detections` and `utils.discard_outliers'
    extent   : float
        parameter controlling the amount of area around each detection bounding box which should be used to generate the volume.
        An extent of x means extent by the square on each side by x*edge_length, where edge_length is the bigger of the width and height of the box
    resampling_mode: str
        mode to use for resampling the scan. Some options are 'nearest', 'linear', 'cubic' and 'bicubic'

    Outputs
    -------
    vert_dicts : list of dictionaries
        A modified version of the input `vert_dicts', except with a new key `volume' added to each dictionary containing the volume
        surrounding the vertebrae as a 224x224x16 np.array
    """

    for vert_idx, vert in enumerate(vert_dicts):
        detection_poly = vert["average_polygon"]
        detection_poly = np.array(detection_poly)

        rotated_scan, new_bb = straighten_bb(scan, detection_poly)

        new_bb_width = new_bb[:, 1].ptp()
        new_bb_height = new_bb[:, 0].ptp()
        edge_len = np.max([new_bb_height, new_bb_width])
        y_min = int(new_bb[-1, 1] - extent * edge_len)
        x_min = int(new_bb[-1, 0] - extent * edge_len)
        y_max = int(y_min + (1 + 2 * extent) * edge_len)
        x_max = int(x_min + (1 + 2 * extent) * edge_len)
        try:
            if y_min < 0:
                diff = -y_min
                diff_patch = np.zeros((diff, x_max - x_min, rotated_scan.shape[2]))
                vol = rotated_scan[0:y_max, x_min:x_max]
                volume = np.concatenate((diff_patch, vol), axis=0)
            else:
                volume = rotated_scan[y_min:y_max, x_min:x_max]

            resized_bb = resize_bb(
                volume,
                output_shape=output_shape,
                only_2d_interpolation=(not rescale_sagittal),
                resampling_mode=resampling_mode,
            ).numpy()

            vert["volume"] = resized_bb
        except Exception as error:
            print(error)
            print("Warning: Could not resize vertebral volume!")
            print(y_min, x_min, x_max, y_max)
            vert["volume"] = np.zeros(output_shape)
    return vert_dicts


def straighten_bb(scan, bounding_box):
    """
    Given a 3d scan and 2d bounding boxes, rotate the boxes and the scan so that the top edge of the bounding box is horizontal and give the coordinates of the bounding box in the new scan

    Inputs
    ------
    scan: np.array
        a 3d numpy tensor of dimensions height, width and depth (sagittal plane)
    bounding_box: np.array
        a 4x2 tensor with with [:,0] containing the x coordinates of the box corners and [:,1] containing the y coordinates
    Outputs
    -------
    rotated_scan: np.array
        a 3d numpy tensor of dimensions height, width and depth (sagittal plane). Rotated version of scan
    new_bounding_box: np.array
        a 4x2 tensor with with [:,0] containing the x coordinates of the box corners and [:,1] containing the y coordinates. Same co-ords as bounding_box, transformed to the frame of rotated_scan
    """
    x_lt = bounding_box[-1, 0]
    y_lt = bounding_box[-1, 1]
    x_rt = bounding_box[0, 0]
    y_rt = bounding_box[0, 1]
    delta_y = y_rt - y_lt
    delta_x = x_rt - x_lt
    theta = np.tanh(delta_y / delta_x) * 180 / np.pi

    rotation_matrix = cv2.getRotationMatrix2D(
        (np.mean([x_lt, x_rt]), np.mean([y_lt, y_rt])), theta, scale=1
    )
    # get new_bounding_box
    new_bounding_box = (
        rotation_matrix @ [bounding_box[:, 0], bounding_box[:, 1], np.ones(4)]
    ).T

    rotated_scan = np.zeros_like(scan).astype(np.float32)
    for i in range(scan.shape[-1]):
        try:
            rotated_scan[:, :, i] = cv2.warpAffine(
                scan[:, :, i],
                rotation_matrix,
                (scan.shape[1], scan.shape[0]),
                flags=cv2.INTER_CUBIC,
            )
        except Exception as e:
            print(e)
            import pdb

            pdb.set_trace()

    return rotated_scan, new_bounding_box


def resize_bb(
    vol,
    output_shape=(224, 224, 16),
    only_2d_interpolation=False,
    resampling_mode="bicubic",
):
    """
    Resizes a 3d tensor to be of size `output_shape` and transforms a bounding box into the new shape

    Inputs
    ------
    vol: np.array
        the 3d tensor to be resized
    output_shape:tuple
        the 3d shape to make interpolate vol to
    Outputs
    -------
    new_vol: np.array
        the new, interpolated volume of size output shape
    """
    orig_vol_shape = vol.shape
    vol = torch.einsum("ijk->kij", torch.tensor(vol)).unsqueeze(1).type(torch.double)
    if resampling_mode in ["bicubic", "linear", "bicubic"]:
        new_vol = F.interpolate(
            vol,
            (output_shape[0], output_shape[1]),
            mode=resampling_mode,
            align_corners=False,
        ).squeeze(1)
    else:
        new_vol = F.interpolate(
            vol,
            (output_shape[0], output_shape[1]),
            mode=resampling_mode,
        ).squeeze(1)

    new_vol = torch.einsum("kij->ijk", new_vol)

    new_new_vol = torch.zeros(output_shape)

    ratio = float(new_vol.shape[-1]) / float(output_shape[-1])
    for i in range(output_shape[-1]):
        interp_val = i * ratio
        high_slice = int(np.ceil(interp_val))
        low_slice = int(np.floor(interp_val))
        if high_slice >= new_vol.shape[-1]:
            high_slice = new_vol.shape[-1] - 1
        if low_slice >= new_vol.shape[-1]:
            low_slice = new_vol.shape[-1] - 1
        if high_slice == low_slice:
            new_new_vol[:, :, i] = new_vol[:, :, high_slice]
        else:
            new_new_vol[:, :, i] = new_vol[:, :, high_slice] * np.abs(
                low_slice - interp_val
            ) + new_vol[:, :, low_slice] * np.abs(high_slice - interp_val)

    arr_percentile = np.percentile(new_new_vol[:, :], 95)
    min_ = new_new_vol.min()
    for slice_ in range(new_new_vol.shape[-1]):
        new_new_vol[:, :, slice_] = (new_new_vol[:, :, slice_] - min_) / (
            arr_percentile + 0.000001
        )

    if only_2d_interpolation:
        new_new_vol = torch.tensor(new_vol)
    # change this to new_new_vol to get 16  slice volumes back
    return new_new_vol



import os
import sys
import glob
import torch
import numpy as np
from scipy import io as sio
from collections import OrderedDict
from pathlib import Path
from typing import List, Tuple, Dict, Union, Any
from typing_extensions import TypedDict


IVDDicts = TypedDict('IVDDicts', {'volume': np.array, 'level_name': str})
VertDicts = TypedDict('VertDicts', {})

sys.path.append(str(Path(__file__).parent))

# Ladder
from .utils import *


class SpineNet:

    def __init__(
        self, device: bool = "cuda:0", verbose: bool = True, scan_type: str = "lumbar"
    ) -> None:
        """
        Initialize instance of spinenet for (1) detecting and labelling vertebrae 
        (2) Performing radiological grading for common spinal degenerative changes in T2 sagittal lumbar scans.

        Parameters
        ----------
        device : str, optional
            The pytorch-style device to use for the model. The default is "cuda:0". If you not using CUDA-enabled machine, you can use "cpu" (although this will slow performance).
        verbose : bool, optional
            Whether to print out information regarding the pipeline. The default is True.
        scan_type : str, optional
            The type of scan to use. The default is "lumbar". Can also be "whole"

        """

        assert scan_type in [
            "lumbar",
            "whole",
        ], "scan_type should be either 'lumbar' or 'whole'"
        self.device = device
        if "cuda" in device:
            torch.cuda.set_device(device)
        # the value of the threshold of corner channels in the detector network
        self.corner_threshold = 0.6
        # the value of the threshold of centroid channels in the detector network
        self.centroid_threshold = 0.6
        # the amount of overlap two polygons need to have to be grouped together
        self.group_across_slices_threshold = 0.2
        self.temp_path = "./temp"  # the path to save temporary files to

        # sets whether to split scan into pieces
        if scan_type == "lumbar":
            self.remove_black_space = False
        else:
            self.remove_black_space = True

        # Setup - Detector Model
        from .models.vfr import VFRResNetDetector

        detection_weights_path = os.path.join(
            os.path.dirname(__file__), "weights/detect-vfr"
        )
        self.detection_model = VFRResNetDetector()
        if verbose:
            print("Loading Detection Model...")
        self.detection_model.load_weights(detection_weights_path, verbose=verbose)
        self.detection_model.eval()
        self.detection_model.to(device)

        # Setup - Appearance Model
        from .models.appearance import AppearanceModel

        appearance_weights_path = os.path.join(
            os.path.dirname(__file__), "weights/appearance"
        )
        self.appearance_model = AppearanceModel()
        if verbose:
            print("Loading Appearance Model...")
        self.appearance_model.load_weights(appearance_weights_path, verbose=verbose)
        self.appearance_model.eval()
        self.appearance_model.to(device)

        # Setup - Context Model
        from .models.context import ContextModel

        context_weights_path = os.path.join(
            os.path.dirname(__file__), "weights/context"
        )
        self.context_model = ContextModel()
        if verbose:
            print("Loading Context Model...")
        self.context_model.load_weights(context_weights_path, verbose=verbose)
        self.context_model.eval()
        self.context_model.to(device)


        # Setup - Grading Model
        from .models.grading import GradingModel

        grading_weights_path = os.path.join(
            os.path.dirname(__file__), "weights/grading"
        )
        self.grading_model = GradingModel()
        if verbose:
            print("Loading Grading Model...")
        self.grading_model.load_weights(grading_weights_path, verbose=verbose)
        self.grading_model.eval()
        self.grading_model.to(device)

    def detect_vb(
        self,
        volume : np.ndarray,
        pixel_spacing : Union[np.ndarray, List[float], torch.Tensor],
        debug: bool = False,
        penalise_skips: bool = True,
        remove_single_slice_detections: bool = True,
    ) -> VertDicts:
        """
        Use SpineNet to detect and label vertebral bodies in a volume.

        Parameters
        ----------
        volume : np.ndarray
            The volume to detect vertebrae in. Should have shape (height,width, number of sagittal slices).
        pixel_spacing : Union[np.ndarray, List[float], torch.Tensor]
            The pixel spacing of the volume, specifically the distance between adjacent pixels in the sagittal direction.
            This has order height, width
        debug : bool, optional
            Whether to print out debug information. The default is False.
        penalise_skips : bool, optional
            Whether to penalise skipping a vertebrae in the detection sequence. Essentially, enabling this ensures the vertebrae labels will be in order (i.e. T12, L5, L4, etc).
            However, this may not be desired if you expect to miss detections (perhaps due to image artifacts)

        Returns
        -------
        VertDicts
            A list of dictionaries containing the vertebrae labels, their corresponding polygons and the slices in which they appear.
        """

        assert volume.ndim == 3, "scan should be a 3-dimensional array of shape HxWxS."
        detect_ans = detect_and_group(
            self.detection_model,
            volume,
            remove_excess_black_space=self.remove_black_space,
            plot_outputs=False,
            using_resnet=True,
            corner_threshold=self.corner_threshold,
            centroid_threshold=self.centroid_threshold,
            group_across_slices_threshold=self.group_across_slices_threshold,
            remove_single_slice_detections=remove_single_slice_detections,
            pixel_spacing=pixel_spacing,
            device=self.device,
            debug=debug,
        )
        if debug:
            vert_dicts, patches, patches_dicts, detection_dicts, transform_info_dicts = detect_ans
        else:
            vert_dicts = detect_ans

        vert_dicts = extract_volumes(volume, vert_dicts)
        labels_ans = label_verts(
            vert_dicts,
            volume,
            pixel_spacing,
            self.appearance_model,
            self.context_model,
            plot_outputs=False,
            penalise_skips=penalise_skips,
            debug=debug,
        )
        if debug:
            (
                vert_dicts,
                height_scaled_appearance_features,
                context_output,
                context_features,
                seq_pred,
            ) = labels_ans
        else:
            vert_dicts = labels_ans

        if debug:
            return (
                vert_dicts,
                patches,
                patches_dicts,
                detection_dicts,
                transform_info_dicts,
                height_scaled_appearance_features,
                context_output,
                context_features,
                seq_pred,
            )
        else:
            return vert_dicts

    def get_ivds_from_vert_dicts(self, vert_dicts : VertDicts, scan_volume: np.ndarray) -> IVDDicts:
        '''
        Use detected vertebrae from SpineNet's detect_vb function to generate volumes surrounding each IVD.

        Parameters
        ----------
        vert_dicts : VertDicts  
            The vertebrae labels, their corresponding polygons and the slices in which they appear. Generated by SpineNet.detect_vb function
        scan_volume : np.ndarray
            The volume to extract the IVDs from. Should have shape (height,width, number of sagittal slices).
        
        Returns
        -------
        IVDDicts
            A list of dictionaries containing resampled IVD volumes and the corresponding names
        '''
        (
            all_vb_x,
            all_vb_y,
            all_vb_mid,
            all_vb_label,
            vb_level_names,
        ) = vert_dicts_to_classification_format(vert_dicts, scan_volume.shape[-1])
        ivds = get_all_ivd_vol(
            scan_volume, all_vb_x, all_vb_y, all_vb_mid, all_vb_label
        )
        ivd_level_names = get_ivd_level_names(vb_level_names)
        ivd_dicts = [{"volume": ivd, "level_name": level_name} for ivd, level_name in zip(ivds, ivd_level_names)]
        return ivd_dicts

    def grade_ivds(self, ivd_dicts : IVDDicts) -> pd.DataFrame:
        """
        Grade all intervertebral discs extracted from a volume. See the technical paper for details on the specific grading schemes: https://arxiv.org/abs/2205.01683.

        Parameters
        ----------
        ivd_dicts : IVDDicts
            A list of dictionaries containing resampled IVD volumes and the corresponding names. Generated by SpineNet.get_ivds_from_vert_dicts function

        Returns
        -------
        gradings: pd.DataFrame
            A pandas dataframe containing the IVD names and their corresponding grades.
        """
        gradings = classify_ivd_v2_resnet(self.grading_model, [ivd_dict['volume'] for ivd_dict in ivd_dicts], self.device)
        gradings = format_gradings(gradings, [ivd_dict['level_name'] for ivd_dict in ivd_dicts])
        return gradings



"""
SpineNet Integration Test Suite

Tests core functionality including scan loading, vertebrae detection,
and results saving. Designed for CI/CD with efficient resource management.
"""
import os
import pytest
import torch
import numpy as np
from pathlib import Path

import spinenet
from spinenet import SpineNet, download_example_scan
from spinenet.io import load_dicoms_from_folder, save_vert_dicts_to_csv


@pytest.fixture(scope="session")
def setup_weights():
    """Download model weights once per test session."""
    spinenet.download_weights(verbose=True, force=False)


@pytest.fixture(scope="session")
def example_scan_folder(tmp_path_factory, setup_weights):
    """Download example scan data once per session."""
    scan_name = 't2_lumbar_scan_1'
    folder = tmp_path_factory.mktemp('example_scans')
    download_example_scan(scan_name, file_path=str(folder))
    return folder / scan_name


@pytest.fixture(scope="session")
def loaded_scan(example_scan_folder):
    """Load scan with metadata overrides for consistent testing."""
    overwrite_dict = {
        'SliceThickness': [2],
        'ImageOrientationPatient': [0, 1, 0, 0, 0, -1]
    }
    scan = load_dicoms_from_folder(
        str(example_scan_folder),
        require_extensions=False,
        metadata_overwrites=overwrite_dict
    )
    return scan


@pytest.fixture(scope="session")
def spinenet_model(setup_weights):
    """Create SpineNet model instance once per session."""
    model = SpineNet(device='cpu', verbose=True, scan_type='lumbar')
    return model


@pytest.fixture(scope="session")
def detection_results(loaded_scan, spinenet_model):
    """Run detection once and cache results for all tests."""
    return spinenet_model.detect_vb(loaded_scan.volume, loaded_scan.pixel_spacing)


class TestScanLoading:
    """Test scan loading and data structure validation."""

    def test_scan_attributes(self, loaded_scan):
        """Test that loaded scan has required attributes."""
        scan = loaded_scan
        assert hasattr(scan, 'volume'), "Scan should have volume attribute"
        assert hasattr(scan, 'pixel_spacing'), "Scan should have pixel_spacing attribute"

    def test_volume_structure(self, loaded_scan):
        """Test volume data structure and dimensionality."""
        volume = loaded_scan.volume
        assert isinstance(volume, np.ndarray), "Volume should be numpy array"
        assert len(volume.shape) == 3, "Volume should be 3D (H x W x D)"
        assert volume.size > 0, "Volume should not be empty"

    def test_pixel_spacing_format(self, loaded_scan):
        """Test pixel spacing data format."""
        spacing = loaded_scan.pixel_spacing
        assert isinstance(spacing, (list, tuple, np.ndarray)), "Pixel spacing should be array-like"
        assert len(spacing) == 2, "Pixel spacing should have 2 elements"
        assert all(isinstance(s, (int, float)) and s > 0 for s in spacing), "Spacing values should be positive numbers"


class TestVertebraeDetection:
    """Test vertebrae detection functionality."""

    def test_detection_output_structure(self, detection_results):
        """Test detection returns properly structured results."""
        assert isinstance(detection_results, list), "Detection should return a list"
        assert len(detection_results) > 0, "Should detect at least one vertebra"

    def test_detection_dictionary_format(self, detection_results):
        """Test each detection result has required fields."""
        for vert_dict in detection_results:
            assert isinstance(vert_dict, dict), "Each detection should be a dictionary"
            assert 'predicted_label' in vert_dict, "Detection should have predicted_label"
            assert 'average_polygon' in vert_dict, "Detection should have average_polygon"

    def test_detection_labels(self, detection_results):
        """Test detection labels are valid strings."""
        labels = [vd['predicted_label'] for vd in detection_results]
        assert len(labels) > 0, "Should have at least one vertebra label"
        assert all(isinstance(l, str) for l in labels), "All labels should be strings"
        assert any(any(c.isalpha() for c in label) for label in labels), "Labels should contain letters"

    def test_polygon_coordinates(self, detection_results, loaded_scan):
        """Test polygon coordinates are within image bounds."""
        h, w, d = loaded_scan.volume.shape
        for vert_dict in detection_results:
            poly = np.array(vert_dict['average_polygon'])
            assert poly.ndim == 2, "Polygon should be 2D array"
            assert poly.shape[1] == 2, "Polygon points should be (x, y) coordinates"
            x_coords, y_coords = poly[:, 0], poly[:, 1]
            assert np.all(x_coords >= 0) and np.all(x_coords < w), "X coordinates should be within image width"
            assert np.all(y_coords >= 0) and np.all(y_coords < h), "Y coordinates should be within image height"


class TestResultsSaving:
    """Test results saving and file I/O functionality."""

    def test_csv_export(self, detection_results, tmp_path):
        """Test saving detection results to CSV file."""
        results_file = tmp_path / "test_vertebrae_detection.csv"
        save_vert_dicts_to_csv(detection_results, str(results_file))
        assert results_file.exists(), "Results file should be created"
        assert results_file.stat().st_size > 0, "Results file should not be empty"

    def test_csv_content_format(self, detection_results, tmp_path):
        """Test CSV file contains expected content structure."""
        results_file = tmp_path / "test_vertebrae_detection.csv"
        save_vert_dicts_to_csv(detection_results, str(results_file))

        # Read and verify CSV content
        content = results_file.read_text()
        lines = content.strip().split('\n')
        assert len(lines) > 1, "CSV should have header and at least one data row"
        assert ',' in lines[0], "CSV should be comma-separated"


class TestIntegration:
    """Integration tests combining multiple components."""

    def test_end_to_end_workflow(self, loaded_scan, spinenet_model, tmp_path):
        """Test complete workflow from scan to saved results."""
        # Run detection
        vert_dicts = spinenet_model.detect_vb(loaded_scan.volume, loaded_scan.pixel_spacing)

        # Save results
        results_file = tmp_path / "integration_test_results.csv"
        save_vert_dicts_to_csv(vert_dicts, str(results_file))

        # Verify workflow success
        assert len(vert_dicts) > 0, "Should detect vertebrae"
        assert results_file.exists(), "Should save results successfully"
        assert results_file.stat().st_size > 0, "Results file should contain data"



import os
import torchvision
import torch
import cv2
from shapely.geometry import Polygon, Point
import torch.nn.functional as F
import os
import numpy as np
import matplotlib.pyplot as plt
import scipy.ndimage
from .scan_preprocessing import split_into_patches_exhaustive, split_into_patches_exhaustive_spacing
from .detection_post_processing import make_in_slice_detections


def detect_and_group(
    detection_net,
    scan,
    remove_excess_black_space=True,
    pixel_spacing=1,
    plot_outputs=False,
    using_resnet=True,
    corner_threshold=0.5,
    centroid_threshold=0.5,
    group_across_slices_threshold=0.2,
    remove_single_slice_detections=True,
    device="cuda:0",
    debug=False,
):
    """
    Detects all vertebrae corners, groups into vertebrae and returns results

    Inputs
    ------
        detection_net : torch.nn.Module
            the VFR model used to detect vertebrae landmarks and corners
        scan : np.array
            A HxWxS array containing the scan
        remove_excess_black_space : bool [optional]
            remove any excess space at the sides of the scan. DEPRECIATED
        plot_outputs : bool [optional]
            Show response maps for detection channels of detection_net output
        using_resnet :bool [optional]
            A flag that tells the function is it is using a resnet model. If
            so this changes the normalization
        centroid_thresholding : float 0<=x<=1
            The value at which to threshold the detector output in the centroid
            channel. Higher values means more detections but risks false
            positives
        corner_thresholding : float 0<=x<=1
            Like the centroid thresholding parameter but for the corner channels
            of the detection network

    Outputs
    -------
        vert_dicts : list of dicts
            A list of dictionaries containing information about the vertebrae
            detected in the scan. Each element of the list corresponds to a
            different vertebra with the key 'average_polygon' containing the
            median location of the vertebrae through the slices, 'slice_nos'
            containing the slices in which the polygon appears and 'polys' the
            polygons detected.
    """


    # split the scan into different patches
    # patches, transform_info_dicts = split_into_patches_exhaustive(
    #     scan, pixel_spacing=pixel_spacing, overlap_param=0.4, using_resnet=using_resnet
    # )
    # split the scan into different patches
    patches, transform_info_dicts = split_into_patches_exhaustive_spacing(
        scan, pixel_spacing=pixel_spacing, overlap_param=0.4, using_resnet=using_resnet
    )
    # group the detections made in each patch into slice level detections
    detection_dicts, patches_dicts = make_in_slice_detections(
        detection_net,
        patches,
        transform_info_dicts,
        scan.shape,
        corner_threshold,
        centroid_threshold,
        device=device,
    )

    vert_dicts = group_slice_detections(
        detection_dicts,
        iou_threshold=group_across_slices_threshold,
        remove_single_slice_detections=remove_single_slice_detections,
    )

    if not debug:
        return vert_dicts
    else:
        return vert_dicts, patches, patches_dicts, detection_dicts, transform_info_dicts


def group_slice_detections(
    detection_dicts, iou_threshold=0.1, remove_single_slice_detections=True
):
    """
    group polygons detected across slices to get 3-d bounding quadrilateral for vertebra

    Inputs
    ------
    detection_dicts: list of dicts
        list of dictionaries of length equal to the number of slices. Each dictionary describes
        the detections made in a given slice. TODO - describe structure of these dictionaries

    Outputs
    -------
    vert_dicts : list
        list of dictionaries describing each vertebra detected. Each element corresponds to a
        different vertebra. TODO - describe structure of these dictionaries
    """
    # TODO: Complete docstring
    vert_dicts = []
    # loop through slices
    for slice_no, slice_detections in enumerate(detection_dicts):
        # loop through vert bodies detected in each slice
        for polygon_in_slice in slice_detections["detection_polys"]:
            # now loop through previous detections and see if any match up
            overlaps_with_previous = False
            for vert in vert_dicts:
                try:
                    most_recent_polygon = vert["polys"][-1]
                    poly_ious = get_poly_iou(polygon_in_slice, most_recent_polygon)
                except:
                    poly_ious = 0
                if poly_ious > iou_threshold:
                    vert["polys"].append(polygon_in_slice)
                    vert["slice_nos"].append(slice_no)
                    # recalculate average poly for that vertebrae
                    vert["average_polygon"] = np.mean(vert["polys"], axis=0)
                    overlaps_with_previous = True
                    break
            # make new entry if doesn't overlap with any of the previous ones
            if overlaps_with_previous is False:
                vert_dicts.append(
                    {
                        "polys": [polygon_in_slice],
                        "average_polygon": polygon_in_slice,
                        "slice_nos": [slice_no],
                    }
                )
    remove_indices = []  # list to store indicies of verts to be deleted
    if remove_single_slice_detections:
        for vert_index, vert in enumerate(vert_dicts):
            if len(vert["polys"]) < 2:
                remove_indices.insert(
                    0, vert_index
                )  # insert at beginning to get reverse order
        [vert_dicts.pop(remove_index) for remove_index in remove_indices]

    # go through all vertebrae and check their average polygons dont overlap
    match_vert_idxs = []
    for vert_idx, vert_dict in enumerate(vert_dicts):
        for other_vert_idx, other_vert_dict in enumerate(vert_dicts):
            if vert_idx <= other_vert_idx:
                continue
            iou = get_poly_iou(
                vert_dict["average_polygon"], other_vert_dict["average_polygon"]
            )
            if iou > 0.05:
                match_vert_idxs.append([vert_idx, other_vert_idx])
                print('merging')

    # join the matching verts
    for matching_vert_pair in match_vert_idxs:
        vert_dicts[matching_vert_pair[0]]["polys"] += vert_dicts[matching_vert_pair[1]][
            "polys"
        ]
        vert_dicts[matching_vert_pair[0]]["slice_nos"] += vert_dicts[
            matching_vert_pair[1]
        ]["slice_nos"]
        vert_dicts[matching_vert_pair[0]]["average_polygon"] = np.mean(
            vert_dicts[matching_vert_pair[0]]["polys"], axis=0
        )
        print(f"merging vert {matching_vert_pair[0]} with {matching_vert_pair[1]}")

    # remove extra verts
    for matching_vert_pair in sorted(match_vert_idxs, key=lambda x: x[1], reverse=True):
        vert_dicts.pop(matching_vert_pair[1])

    # now sort list in height order
    vert_dicts.sort(key=lambda x: np.mean(np.array(x["average_polygon"])[:, 1]))
    return vert_dicts


def get_poly_iou(poly1, poly2):
    """
    calculates and returns IOU of two polygons
    Inputs
    ------
    poly1 : np.array
        4x2  numpy array containing vertices of quadrilateral
    poly2 : np.array
        4x2  numpy array containing vertices of quadrilateral

    Outputs
    -------
    iou : float
        the iou of poly1 and poly2
    """

    poly1 = Polygon(poly1)
    poly2 = Polygon(poly2)
    intersection = poly1.intersection(poly2).area
    union = poly1.union(poly2).area
    min_area = np.min([poly1.area, poly2.area])
    if union > 0:
        return intersection / union
    else:
        return 0


def red(x):
    y = np.zeros_like(x)
    return np.stack([x, y, y], axis=-1)


def blue(x):
    y = np.zeros_like(x)
    return np.stack([y, y, x], axis=-1)


def green(x):
    y = np.zeros_like(x)
    return np.stack([y, x, y], axis=-1)


def yellow(x):
    y = np.zeros_like(x)
    return np.stack([x, x, y], axis=-1)


def pink(x):
    y = np.zeros_like(x)
    return 0.5 * (yellow(x) + red(x))


def color(x):
    y = np.zeros_like(x)
    return np.stack([x, x, x], axis=-1)



import os
from typing import List, Tuple, Dict, Union, Any
import zipfile

from .io import download_weights as download_weights_io, download as download_io
from .main import SpineNet
from .utils import *


def download_weights(verbose: bool = False, force: bool = False) -> None:
    """
    Download SpineNet model weights from public server and save to a location
    where the pipeline expects them (`./spinenet/weights/`).

    Parameters
    ----------
    verbose : bool, optional
        Whether to print out extra information about the file downloads.
    force : bool, optional
        Set to force download weights even if they already have been locally downloaded
    """

    weights_dict = {
        "grading/ckpt1.pt": "https://www.robots.ox.ac.uk/~vgg/research/spinenet/weights/grading/ckpt1.pt",
        "appearance/ckpt187.pt": "https://www.robots.ox.ac.uk/~vgg/research/spinenet/weights/appearance/ckpt187.pt",
        "detect-vfr/ckpt435.pt": "https://www.robots.ox.ac.uk/~vgg/research/spinenet/weights/detect-vfr/ckpt435.pt",
        "context/ckpt16.pt": "https://www.robots.ox.ac.uk/~vgg/research/spinenet/weights/context/ckpt16.pt"
    }
    weights_path = os.path.join(os.path.dirname(__file__),"weights")

    for path, url in weights_dict.items():
        if not os.path.exists(os.path.join(weights_path, os.path.dirname(path))):
            os.makedirs(os.path.join(weights_path, os.path.dirname(path)), exist_ok=True)

    if verbose:
        print("Downloading weights...")
    download_weights_io(weights_path, weights_dict, verbose=verbose, force=force)

def download_example_scan(example_scan_name: str, file_path: Union[os.PathLike, str]) -> None:
    """
    Download & unzip example scan DICOM folders from public server for testing and tutorials.
    The scans are mostly taken from radiopedia.org and are
    attributed in the project README.md.

    Parameters
    ----------
    example_scan_name : str
        The example scan name to download locally from the server. Current options are
        't2_lumbar_scan_1', 't2_lumbar_scan_2', 'stir_whole_spine'
    file_path
        the directory to save the scan name to. For example 't2_lumbar_scan_1' would save to
        file_path/t2_lumbar_scan_1
    """
    scans_path = {
        't2_lumbar_scan_1': 'https://www.robots.ox.ac.uk/~vgg/research/spinenet/example_scans/t2_lumbar_scan_1.zip',
        't2_lumbar_scan_2': 'https://www.robots.ox.ac.uk/~vgg/research/spinenet/example_scans/t2_lumbar_scan_2.zip',
        'stir_whole_spine': 'https://www.robots.ox.ac.uk/~vgg/research/spinenet/example_scans/stir_whole_spine.zip',
        't2_whole_spine': 'https://www.robots.ox.ac.uk/~vgg/research/spinenet/example_scans/t2_whole_spine.zip',
    }
    assert example_scan_name in scans_path.keys(), f"example_scan_name must be one of {scans_path.keys()}"



    scan_url = scans_path[example_scan_name]

    download_io(scan_url, os.path.join(file_path, 'temp_file.zip'))

    # Unzip the file
    with zipfile.ZipFile(os.path.join(file_path,'temp_file.zip'), 'r') as zip_ref:
        zip_ref.extractall(file_path)

    # Remove the temp file
    os.remove(os.path.join(file_path,'temp_file.zip'))





import torch
import torch.nn as nn
import torch.nn.functional as F
import os
import glob


class AppearanceModel(nn.Module):
    def __init__(self):
        """
        conv dropout chooses if there is dropout between the conv layers
        """
        super(AppearanceModel, self).__init__()
        self.conv1 = nn.Conv3d(1, 8, (3, 3, 1), padding=(3, 3, 0), stride=(2, 2, 1))
        self.bnrm1 = nn.BatchNorm3d(8)
        self.conv2 = nn.Conv3d(8, 16, (3, 3, 1), padding=(3, 3, 0))
        self.bnrm2 = nn.BatchNorm3d(16)
        self.conv3 = nn.Conv3d(16, 32, (3, 3, 3))
        self.bnrm3 = nn.BatchNorm3d(32)
        self.conv4 = nn.Conv3d(32, 64, (3, 3, 3))
        self.bnrm4 = nn.BatchNorm3d(64)
        self.conv5 = nn.Conv3d(64, 128, (3, 3, 3))
        self.bnrm5 = nn.BatchNorm3d(128)
        self.fc1 = nn.Linear(6400, 256)
        self.fc2 = nn.Linear(256, 24)
        self.fc_dropout_layer = nn.Dropout()

    def forward(self, x):
        x = self.get_appearance_features(x)
        x = self.fc_dropout_layer(x)
        x = self.fc2(x)
        return x

    def get_appearance_features(self, x):
        x = self.bnrm1(F.relu(self.conv1(x)))
        x = F.max_pool3d(self.bnrm2(F.relu(self.conv2(x))), kernel_size=(2, 2, 1))
        x = F.max_pool3d(self.bnrm3(F.relu(self.conv3(x))), kernel_size=(2, 2, 1))
        x = F.max_pool3d(self.bnrm4(F.relu(self.conv4(x))), kernel_size=2)
        x = F.max_pool3d(self.bnrm5(F.relu(self.conv5(x))), kernel_size=2)

        # flatten layers
        x = x.flatten(start_dim=1)
        x = self.fc_dropout_layer(x)
        x = F.relu(self.fc1(x))
        return x

    def load_weights(self, save_path, verbose=True):
        if os.path.isdir(save_path):
            list_of_pt = [
                x for x in glob.glob(save_path + "/*.pt") if "encrypted" not in x
            ]
            latest_pt = max(list_of_pt, key=os.path.getctime)
            checkpoint = torch.load(latest_pt, map_location="cpu")
            self.load_state_dict(checkpoint["net"])
            best_loss = checkpoint["loss"]
            start_epoch = checkpoint["epoch"] + 1
            if verbose:
                print(f"==> Loading model trained for {start_epoch} epochs...")
        else:
            raise NameError(f"save path {save_path} could not be found")
        return



from .dicom_io import load_dicoms, SpinalScan, load_dicoms_from_folder
from .download import download, download_weights
from .save_results import save_vert_dicts_to_csv

__all_exports = [load_dicoms, load_dicoms_from_folder, save_vert_dicts_to_csv]

for e in __all_exports:
    e.__module__ = __name__

__all__ = [e.__name__ for e in __all_exports]






import numpy as np
from sklearn.metrics import balanced_accuracy_score, accuracy_score


def balanced_accuracy(y_label, y_pred):
    # Move to cpu
    y_pred = np.array(y_pred, dtype=float)
    y_label = np.array(y_label, dtype=float)
    # Delete ignore_index = -100
    y_pred = np.delete(y_pred, np.where(y_label == -100))
    y_label = np.delete(y_label, np.where(y_label == -100))
    return balanced_accuracy_score(y_label, y_pred)


# binarized score before determining balanced accuracy e.g. for normal vs abnormal accuracy
def binarized_balanced_accuracy(y_label, y_pred):
    y_pred = np.array(y_pred, dtype=float)
    y_label = np.array(y_label, dtype=float)
    y_label[np.where(y_label > 1)] = 1
    y_pred[np.where(y_pred > 1)] = 1
    # Delete ignore_index = -100
    y_pred = np.delete(y_pred, np.where(y_label == -100))
    y_label = np.delete(y_label, np.where(y_label == -100))
    return balanced_accuracy_score(y_label, y_pred)
